################################################################################
#
# Copyright (C) 2016-2022 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

from .SolutionStructs import Solution
from .Common import printExit, restoreDefaultGlobalParameters, assignGlobalParameters, ensurePath
from .Utilities.ConditionalImports import yamlDumper
from . import LibraryIO

from copy import deepcopy
from enum import IntEnum
import yaml
import os
import sys
import argparse

verbosity = 1

class Logic:
    def __init__(self, sols, sizes):
        self.sols = sols
        self.sizes = sizes

def allFiles(startDir):
    current = os.listdir(startDir)
    files = []
    for filename in [_current for _current in current if os.path.splitext(_current)[-1].lower() == '.yaml']:
        fullPath = os.path.join(startDir,filename)
        if os.path.isdir(fullPath):
            files = files + allFiles(fullPath)
        else:
            files.append(fullPath)
    return files

def reindexSolutions(logic):
    for i, s in enumerate(logic.sols):
        s["SolutionIndex"] = i

def fixSizeInconsistencies(sizes, fileType):
    duplicates = list()
    for i in range(0,len(sizes)):
        currSize = sizes[i][0]
        # >= so size will be trimmed when a SolutionTag is included
        if len(currSize) >= 8:
            currSize = currSize[:-4]
            if currSize in (item for index in sizes for item in index):
                duplicates.append(i-len(duplicates))
            else:
                sizes[i][0] = currSize
    sizes_ = deepcopy(sizes)
    if len(duplicates) > 0:
        for i in duplicates:
            sizes_.pop(i)
        verbose(len(duplicates), "duplicate size(s) removed from", fileType, "logic file")
    return sizes_, len(sizes_)

def solsEq(s1o, s2o):
    s1 = deepcopy(s1o)
    s2 = deepcopy(s2o)

    s1["AssignedProblemIndependentDerivedParameters"] = False
    s1["AssignedDerivedParameters"] = False

    s2["AssignedProblemIndependentDerivedParameters"] = False
    s2["AssignedDerivedParameters"] = False

    return Solution(s1) == Solution(s2)

def addKernel(solutionPool, solution):
    for item in solutionPool:
        if solsEq(item, solution):
            index = item["SolutionIndex"]
            debug("...Reuse previously existed kernel", end="")
            break
    else:
        index = len(solutionPool)
        _solution = deepcopy(solution) # if we don't we will see some subtle errors
        _solution["SolutionIndex"] = index
        solutionPool.append(_solution)
        debug("...A new kernel has been added", end="")
    debug("({}) {}".format(index, solutionPool[index]["SolutionNameMin"] if "SolutionNameMin" in solutionPool[index] else "(SolutionName N/A)"))
    return solutionPool, index

def removeUnusedKernels(origLogic, prefix=""):
    origNumSolutions = len(origLogic.sols)

    kernelsInUse = [ index for _, [index, _] in origLogic.sizes ]
    for i, solution in enumerate(origLogic.sols):
        solutionIndex = solution["SolutionIndex"]
        origLogic.sols[i]["__InUse__"] = True if solutionIndex in kernelsInUse else False

    # debug prints
    for o in [o for o in origLogic.sols if o["__InUse__"]==False]:
        debug("{}Solution ({}) {} is unused".format(
            prefix,
            o["SolutionIndex"],
            o["SolutionNameMin"] if "SolutionNameMin" in o else "(SolutionName N/A)"))

    # filter out dangling kernels
    origLogic.sols = [ {k: v for k, v in o.items() if k != "__InUse__"}
                    for o in origLogic.sols if o["__InUse__"]==True ]

    # reindex solutions
    idMap = {} # new = idMap[old]
    for i, solution in enumerate(origLogic.sols):
        idMap[solution["SolutionIndex"]] = i
        origLogic.sols[i]["SolutionIndex"] = i
    for i, [size, [oldSolIndex, eff]] in enumerate(origLogic.sizes):
        origLogic.sizes[i] = [size, [idMap[oldSolIndex], eff]]

    numInvalidRemoved = origNumSolutions - len(origLogic.sols)
    return numInvalidRemoved

# this is for complying the behavior of legacy merge script, where incremental logic
# file always replaces the base logic file even it's slower in performance -
# in the future we may let default force merge policy = False
def defaultForceMergePolicy(incFile):
    if "arcturus" in incFile:
        forceMerge = False
    else:
        forceMerge = True

    return forceMerge

def msg(*args, **kwargs):
    for i in args: print(i, end=" ")
    print(**kwargs)

def verbose(*args, **kwargs):
    if verbosity < 1: return
    msg(*args, **kwargs)

def debug(*args, **kwargs):
    if verbosity < 2: return
    msg(*args, **kwargs)

# Tags distinguishing solution types
# Can be added to size key to allow solutions of each type to be present
# in logic file for a given size
class MfmaTag(IntEnum):
    VALU = 0
    MFMA = 1

    def __str__(self):
        return ["VALU", "MFMA"][self]
    def __repr__(self):
        return str(self)

class AlphaValueTag(IntEnum):
    ANY    = 0
    ONE    = 1
    NEG_ONE = 2
    ZERO   = 3

    def __str__(self):
        return "Alpha="+["Any", "1", "-1", "0"][self]
    def __repr__(self):
        return str(self)

class BetaValueTag(IntEnum):
    ANY    = 0
    ONE    = 1
    NEG_ONE = 2
    ZERO   = 3

    def __str__(self):
        return "Beta="+["Any", "1", "-1", "0"][self]
    def __repr__(self):
        return str(self)

def strToScalarValueTag(Class, value):
    if value == "Any":
        return Class.ANY
    if value == 1:
        return Class.ONE
    if value == -1:
        return Class.NEG_ONE
    if value == 0:
        return Class.ZERO
    else:
        raise RuntimeError("Unsupported value for Alpha/Beta scalar value")

class CEqualsDTag(IntEnum):
    C_EQ_D  = 0
    C_NEQ_D = 1

    def __str__(self):
        return ["C=D", "C!=D"][self]
    def __repr__(self):
        return str(self)

# Tag of form (MFMATag, AlphaValueTag, BetaValueTag, CEqualsDTag)
def getSolutionTag(solution):
    tagTuple = ()
    if solution.get("EnableMatrixInstruction", False) or solution.get("MatrixInstruction", False):
        tagTuple = tagTuple + (MfmaTag.MFMA,)
    else:
        tagTuple = tagTuple + (MfmaTag.VALU,)

    tagTuple = tagTuple + (strToScalarValueTag(AlphaValueTag, solution.get("AssertAlphaValue", "Any")),)
    tagTuple = tagTuple + (strToScalarValueTag(BetaValueTag, solution.get("AssertBetaValue",  "Any")),)

    tagTuple = tagTuple + (CEqualsDTag.C_EQ_D if solution.get("AssertCEqualsD", False) else CEqualsDTag.C_NEQ_D ,)

    return tagTuple

def findSolutionWithIndex(solutionData, solIndex):
    # Check solution at the index corresponding to solIndex first
    if solIndex < len(solutionData) and solutionData[solIndex]["SolutionIndex"] == solIndex:
        return solutionData[solIndex]
    else:
        debug("Searching for index...")
        solution = [s for s in solutionData if s["SolutionIndex"]==solIndex]
        assert(len(solution) == 1)
        return solution[0]

def addSolutionTagToKeys(solutionMap, solutionPool):
    return [[[getSolutionTag(findSolutionWithIndex(solutionPool, idx))] + keys, [idx, eff]]
            for [keys, [idx, eff]] in solutionMap]

def removeSolutionTagFromKeys(solutionMap):
    return [[keys[1:], [idx, incEff]] for keys, [idx, incEff] in solutionMap]

# To be used with add_solution_tags to allow faster general solutions to supercede slower specific ones
def findFastestCompatibleSolution(origDict, sizeMapping):
    tags = sizeMapping[0]
    # Tag of form (MFMATag, AlphaValueTag, BetaValueTag, CEqualsDTag)
    compatibleTagList = [tags]

    # Add all compatible tags to the list
    if tags[1] != AlphaValueTag.ANY:
        compatibleTagList = compatibleTagList + [(t[0], AlphaValueTag.ANY) + t[2:] for t in compatibleTagList]
    if tags[2] != BetaValueTag.ANY:
        compatibleTagList = compatibleTagList + [t[:2] + (BetaValueTag.ANY,) + t[3:] for t in compatibleTagList]
    if tags[3] != CEqualsDTag.C_NEQ_D:
        compatibleTagList = compatibleTagList + [t[:3] + (CEqualsDTag.C_NEQ_D,) + t[4:] for t in compatibleTagList]

    #Find the fastest efficiency of all compatible tags
    maxEfficiency = 0
    for tag in compatibleTagList:
        result = origDict.get((tag,) + sizeMapping[1:], None)
        if result:
            _, eff = origDict[(tag,) + sizeMapping[1:]]
            maxEfficiency = max(maxEfficiency, eff)

    return maxEfficiency

# returns merged logic data as list
def mergeLogic(origLogic, incLogic, forceMerge, trimSize=True, addSolutionTags=False):
    origNumSizes = len(origLogic.sizes)
    origNumSolutions = len(origLogic.sols)

    incNumSizes = len(incLogic.sizes)
    incNumSolutions = len(incLogic.sols)

    verbose(origNumSizes, "sizes and", origNumSolutions, "kernels in base logic file")
    verbose(incNumSizes, "sizes and", incNumSolutions, "kernels in incremental logic file")

    # Add SolutionTag to distinguish solutions with different requirements
    origTaggedSizes = addSolutionTagToKeys(origLogic.sizes, origLogic.sols)
    incTaggedSizes  = addSolutionTagToKeys(incLogic.sizes,  incLogic.sols)
    if addSolutionTags:
        origLogic.sizes = origTaggedSizes
        incLogic.sizes  = incTaggedSizes
    # Print warning if addSolutionTags=False results in removed sizes
    else:
        origSet       = {tuple(size) for size, [_, _] in origLogic.sizes}
        origTaggedSet = {tuple(size) for size, [_, _] in origTaggedSizes}
        incSet        = {tuple(size) for size, [_, _] in incLogic.sizes}
        incTaggedSet  = {tuple(size) for size, [_, _] in incTaggedSizes}

        if len(origSet) != len(origTaggedSet):
            verbose("Warning:", len(origTaggedSet) - len(origSet), "duplicate sizes are present in base logic",
                    "that may not be handled correctly unless --add_solution_tags is used")
        if len(incSet) != len(incTaggedSet):
            verbose("Warning:", len(incTaggedSet) - len(incSet), "duplicate sizes are present in incremental logic",
                    "that may not be handled correctly unless --add_solution_tags is used")

    if trimSize:
        # trim 8-tuple gemm size format to 4-tuple [m, n, b, k]
        # TODO future gemm size could include dictionary format so need robust preprocessing
        [origLogic.sizes, origNumSizes] = fixSizeInconsistencies(origLogic.sizes, "base")
        [incLogic.sizes, incNumSizes] = fixSizeInconsistencies(incLogic.sizes, "incremental")

    numOrigRemoved = removeUnusedKernels(origLogic, "Base logic file: ")
    numIncRemoved = removeUnusedKernels(incLogic, "Inc logic file: ")

    solutionPool = deepcopy(origLogic.sols)
    solutionMap = deepcopy(origLogic.sizes)

    origDict = {tuple(origSize): [i, origEff] for i, [origSize, [origIndex, origEff]] in enumerate(origLogic.sizes)}
    for incSize, [incIndex, incEff] in incLogic.sizes:
        incSolution = findSolutionWithIndex(incLogic.sols, incIndex)

        try:
            j, origEff = origDict[tuple(incSize)]
            if incEff > origEff or forceMerge:
                if incEff > origEff:
                    verbose("[O]", incSize, "already exists and has improved in performance.", end="")
                elif forceMerge:
                    verbose("[!]", incSize, "already exists but does not improve in performance.", end="")
                verbose("Efficiency:", origEff, "->", incEff, "(force_merge=True)" if forceMerge else "")
                solutionPool, index = addKernel(solutionPool, incSolution)
                solutionMap[j][1] = [index, incEff]
            else:
                verbose("[X]", incSize, "already exists but does not improve in performance.", end="")
                verbose("Efficiency:", origEff, "->", incEff)
        except KeyError:
            if addSolutionTags and findFastestCompatibleSolution(origDict, tuple(incSize)) > incEff:
                verbose("[X]", incSize, "has been rejected because a compatible solution already exists with higher performance")
            else:
                verbose("[-]", incSize, "has been added to solution table, Efficiency: N/A ->", incEff)
                solutionPool, index = addKernel(solutionPool, incSolution)
                solutionMap.append([incSize,[index, incEff]])

    verbose(numOrigRemoved, "unused kernels removed from base logic file")
    verbose(numIncRemoved, "unused kernels removed from incremental logic file")

    # Remove SolutionTag for yaml output
    if addSolutionTags:
        solutionMap = removeSolutionTagFromKeys(solutionMap)

    mergedLogic = Logic(solutionPool, solutionMap)
    numReplaced = removeUnusedKernels(mergedLogic, "Merged data: ")

    numSizesAdded = len(solutionMap)-len(origLogic.sizes)
    numSolutionsAdded = len(solutionPool)-len(origLogic.sols)
    numSolutionsRemoved = numReplaced+numOrigRemoved # incremental file not counted

    return [mergedLogic, numSizesAdded, numSolutionsAdded, numSolutionsRemoved]

def getSizesAndSols(data):
    if type(data) is list:
        return Logic(data[5], data[7])
    elif type(data) is dict:
        sols = data["Solutions"]
        sizes = data["Library"]["table"]
        return Logic(sols, sizes)
    else:
        printExit("Bad data")

def updateSizesAndSols(data, logic):
    if type(data) is list:
        data[5] = logic.sols
        data[7] = logic.sizes
    elif type(data) is dict:
        data["Solutions"] = logic.sols
        data["Library"]["table"] = logic.sizes
    else:
        printExit("Bad data")

def avoidRegressions(originalDir, incrementalDir, outputPath, forceMerge, trimSize=True, addSolutionTags=False):
    originalFiles = allFiles(originalDir)
    incrementalFiles = allFiles(incrementalDir)
    ensurePath(outputPath)

    # filter the incremental logic files that have the corresponding base file
    incrementalFiles = [ i for i in incrementalFiles
                         if os.path.split(i)[-1] in [os.path.split(o)[-1] for o in originalFiles] ]

    for incFile in incrementalFiles:
        basename = os.path.split(incFile)[-1]
        origFile = os.path.join(originalDir, basename)
        forceMerge = defaultForceMergePolicy(incFile) if forceMerge is None else forceMerge

        msg("Base logic file:", origFile, "| Incremental:", incFile, "| Merge policy: %s"%("Forced" if forceMerge else "Winner"), "| Trim size:", trimSize,
        "| Add solution tags:", addSolutionTags)

        origData = LibraryIO.readYAML(origFile)
        incData = LibraryIO.readYAML(incFile)

        origLogic = getSizesAndSols(origData)
        incLogic = getSizesAndSols(incData)

        # So far "SolutionIndex" in logic yamls has zero impact on actual 1-1 size mapping (but the order of the Solution does)
        # since mergeLogic() takes that value very seriously so we reindex them here so it doesn't choke on duplicated SolutionIndex
        reindexSolutions(origLogic)
        reindexSolutions(incLogic)

        mergedLogic, *stats = mergeLogic(origLogic, incLogic, forceMerge, trimSize, addSolutionTags)
        msg(stats[0], "size(s) and", stats[1], "kernel(s) added,", stats[2], "kernel(s) removed")

        updateSizesAndSols(origData, mergedLogic)

        with open(os.path.join(outputPath, basename), "w") as outFile:
            yaml.dump(origData, outFile, yamlDumper, default_flow_style=None)
        msg("File written to", os.path.join(outputPath, basename))
        msg("------------------------------")

def main():
    restoreDefaultGlobalParameters()
    assignGlobalParameters({})

    argParser = argparse.ArgumentParser()
    argParser.add_argument("original_dir", help="The library logic directory without tuned sizes")
    argParser.add_argument("incremental_dir", help="The incremental logic directory")
    argParser.add_argument("output_dir", help="The output logic directory")
    argParser.add_argument("-v", "--verbosity", help="0: summary, 1: verbose, 2: debug", default=1, type=int)
    argParser.add_argument("--force_merge", help="Merge previously known sizes unconditionally. Default behavior if not arcturus", default="none")
    argParser.add_argument("--notrim", help="Do not trim long size format down to short format (m,n,b,k). Default is --trim", action="store_false")
    argParser.add_argument("--add_solution_tags", help="Add tags to the size key for solution properies, allowing for solutions with different requirements "
                           "to exist for the same size. Default doesn't add this tag.", action="store_true")

    args = argParser.parse_args(sys.argv[1:])
    originalDir = args.original_dir
    incrementalDir = args.incremental_dir
    outputPath = args.output_dir
    global verbosity
    verbosity = args.verbosity
    forceMerge = args.force_merge.lower()
    trimSize = args.notrim
    add_solution_tags = args.add_solution_tags

    if forceMerge in ["none"]: forceMerge=None
    elif forceMerge in ["true", "1"]: forceMerge=True
    elif forceMerge in ["false", "0"]: forceMerge=False

    avoidRegressions(originalDir, incrementalDir, outputPath, forceMerge, trimSize, add_solution_tags)



################################################################################
#
# Copyright (C) 2016-2024 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

import cProfile
import pstats
import os

from pathlib import Path
from datetime import datetime, timezone
from typing import Callable, Tuple

PROFILE_ENV_VAR = "TENSILE_PROFILE"

def profile(func: Callable) -> Callable:
    """Profiling decorator.

    Add ``@profile`` to mark a function for profiling; set the environment variable
    TENSILE_PROFILE=ON to enable profiling decorated functions.
    """
    if not envVariableIsSet(PROFILE_ENV_VAR):
        return func
    def wrapper(*args, **kwargs):
        path, filename = initProfileArtifacts(func.__name__)

        prof = cProfile.Profile()
        output = prof.runcall(func, *args, **kwargs)
        result = pstats.Stats(prof)
        result.sort_stats(pstats.SortKey.TIME)
        result.dump_stats(path/filename)

        return output
    return wrapper

def envVariableIsSet(varName: str) -> bool:
    """Checks if the provided environment variable is set to "ON", "TRUE", or "1"
    Args:
        varName: Environment variable name.
    Returns:
        True if the environment variable is set, otherwise False.
    """
    value = os.environ.get(varName, "").upper()
    return True if value in ["ON", "TRUE", "1"] else False

def initProfileArtifacts(funcName: str) -> Tuple[Path, str]:
    """Initializes filenames and paths for profiling artifacts based on the current datetime
    Args:
        funcName: The name of the function being profiled, nominally passed via func.__name__
    Returns:
        A tuple (path, filename) where the path is the artifact directory and filename is
        a .prof file with the profiling results.
    """
    dt = datetime.now(timezone.utc)
    filename = f"{funcName}-{dt.strftime('%Y-%m-%dT%H-%M-%SZ')}.prof"
    path = Path().cwd()/f"profiling-results-{dt.strftime('%Y-%m-%d')}"
    path.mkdir(exist_ok=True)
    print(f"> Profiling report at: {str(path / filename)}")
    return path, filename



################################################################################
#
# Copyright (C) 2024 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

import Tensile.SolutionLibrary as SolutionLibrary

from pytest import raises

def test_ArchitectureMap():
    msl = SolutionLibrary.MasterSolutionLibrary(None, None)
    assert msl.ArchitectureIndexMap("gfx90a") == (int("90a", 16) << 18), "Incorrect index value."
    assert msl.ArchitectureIndexMap("gfx900") == (int("900", 16) << 18), "Incorrect index value."
    assert msl.ArchitectureIndexMap("gfx908") == (int("908", 16) << 18), "Incorrect index value."
    assert msl.ArchitectureIndexMap("fallback") == (0), "Incorrect index value."

    # Should throw when computing index for previously computed architecture.
    with raises(RuntimeError, match="ERROR in architecture solution index mapping."):
        msl.ArchitectureIndexMap("gfx90a")

    # Should throw when computing index for non-gfx architecture accept for fallback.
    with raises(RuntimeError, match="ERROR in architecture solution index mapping."):
        msl.ArchitectureIndexMap("test90a")

class MockSolution:
    index = 0
    def __init__(self, name):
        self.name = name

class MockLibrary:
    def remapSolutionIndices(self, indexMap):
        pass

def test_remapSolutionIndexStartingFrom():
    mockSolutions = { 0 : MockSolution("foo"), 1 : MockSolution("bar")}
    SolutionLibrary.MasterSolutionLibrary(mockSolutions, MockLibrary()).remapSolutionIndicesStartingFrom(10)
    assert mockSolutions[0].index == 10, "Zeroith entry should have a remapped index of 10"
    assert mockSolutions[1].index == 11, "First entry should have a remapped index of 11"



################################################################################
#
# Copyright (C) 2020-2022 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

from .Common import globalParameters
from Tensile.Utilities.ConditionalImports import yamlLoader

import yaml

import os

def isCustomKernelConfig(config):
    return "CustomKernelName" in config and config["CustomKernelName"]

def getCustomKernelFilepath(name, directory=globalParameters["CustomKernelDirectory"]):
    return os.path.join(directory, (name + ".s"))

def getAllCustomKernelNames(directory=globalParameters["CustomKernelDirectory"]):
    return [fname[:-2] for fname in os.listdir(directory) if fname.endswith(".s")]

def getCustomKernelContents(name, directory=globalParameters["CustomKernelDirectory"]):
    try:
        with open(getCustomKernelFilepath(name, directory)) as f:
            return f.read()
    except:
        raise RuntimeError("Failed to find custom kernel: {}".format(os.path.join(directory, name)))

def getCustomKernelConfigAndAssembly(name, directory=globalParameters["CustomKernelDirectory"]):
    contents  = getCustomKernelContents(name, directory)
    config = "\n"    #Yaml configuration properties
    assembly = ""
    inConfig = False
    for line in contents.splitlines():
        if   line == "---": inConfig = True                          #Beginning of yaml section
        elif line == "...": inConfig = False                         #End of yaml section
        elif      inConfig: config   += line + "\n"
        else              : assembly += line + "\n"; config += "\n"  #Second statement to keep line numbers consistent for yaml errors

    return (config, assembly)

def getCustomKernelConfig(name, directory=globalParameters["CustomKernelDirectory"]):
    rawConfig, _ = getCustomKernelConfigAndAssembly(name, directory)
    try:
        config = yaml.load(rawConfig, yamlLoader)["custom.config"]
        if "ISA" not in config:
            raise RuntimeError(f"ISA not found in custom kernel config {name}")
        config["ISA"] = tuple(config["ISA"])
        return config
    except Exception as e:
        raise RuntimeError("Failed to read configuration for custom kernel: {0}\nDetails:\n{1}".format(name, e))



################################################################################
#
# Copyright (C) 2020-2022 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

from Tensile.Utilities.String import splitDelimitedString
from Tensile.Utilities.toFile import toFile
from pathlib import Path
import pytest
import os

def test_splitDelimitedString():
    archs = "all"
    expected = {"all"}
    result = splitDelimitedString(archs, {";", "_"})
    assert result == expected, f"arch `{archs}` should parse to {expected} but instead maps to {result}"

    archs = "gfx000;gfx803;gfx900:xnack-"
    expected = {'gfx000', 'gfx803', 'gfx900:xnack-'}
    result = splitDelimitedString(archs, {";", "_"})
    assert result == expected, f"arch `{archs}` should map to {expected} but instead maps to {result}"

    archs = "gfx000_gfx803_gfx900:xnack-"
    expected = {'gfx000', 'gfx803', 'gfx900:xnack-'}
    result = splitDelimitedString(archs, {";", "_"})
    assert result == expected, f"arch `{archs}` should map to {expected} but instead maps to {result}"

    archs = "gfx803,    gfx906_gfx942:gfx1102"
    expected = {"gfx803,    gfx906", "gfx942:gfx1102"}
    result = splitDelimitedString(archs, {";", "_"})
    assert result == expected, f"arch `{archs}` should map to {expected} but instead maps to {result}"

    archs = "gfx900;gfx90a:xnack+-gfx1010"
    expected = {"gfx900", "gfx90a:xnack+-gfx1010"}
    result = splitDelimitedString(archs, {";", "_"})
    assert result == expected, f"arch `{archs}` should map to {expected} but instead maps to {result}"

    archs = ";gfx803;gfx906;"
    expected = {"", "gfx803", "gfx906", ""}
    result = splitDelimitedString(archs, {";", "_"})
    assert result == expected, f"arch `{archs}` should map to {expected} but instead maps to {result}"

    archs = "_gfx803_gfx906_"
    expected = {"", "gfx803", "gfx906", ""}
    result = splitDelimitedString(archs, {";", "_"})
    assert result == expected, f"arch `{archs}` should map to {expected} but instead maps to {result}"

    archs = "abc;gfx90Z;all"
    expected = {"abc", "gfx90Z", "all"}
    result = splitDelimitedString(archs, {";", "_"})
    assert result == expected, f"arch `{archs}` should map to {expected} but instead maps to {result}"

def test_toFile():

    manifest: Path = Path.cwd() / "my-manifest.txt"
    metaData = ["mylib.yaml"]
    codeObjectFiles = ["library/foo.co", "library/bar.co"]
    sourceCodeObjectFiles = ["library/foo.hsaco", "library/bar.hsaco"]

    if manifest.is_file():
        os.remove(manifest)

    with pytest.raises(AssertionError, match="contents must be a list."):
        toFile(manifest, (1,2,3))

    with pytest.raises(AssertionError, match="contents elements must be a str."):
        toFile(manifest, [1,2,3])

    toFile(manifest, metaData + codeObjectFiles + sourceCodeObjectFiles)

    assert manifest.is_file(), "{manifest} was not generated"
    with open(manifest, "r") as f:
        result = f.readlines()

    assert len(result) == 5, "Expected five entries in manifest file."

    if manifest.is_file():
        os.remove(manifest)



################################################################################
#
# Copyright (C) 2021-2024 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

from ..Component import LocalRead
from .. import Code
from ..AsmUtils import vgpr, sgpr
from math import ceil

class LocalReadVALU(LocalRead):
    kernel = {"EnableMatrixInstruction": False}

    """
    Local Read: Do It A/B
    iui = Inner Unroll Idx
    epsi = expand pointer swap index. Only used for PAP
    """
    def __call__(self, writer, bufferIdx, iui, epsi, tP):
        kernel = writer.kernel

        writer.localReadDoCnt += 1

        tc                = tP["tensorChar"]
        tile01            = tP["tile01Idx"]
        imod              = Code.Module("LocalReadDo%s_I%s"%(tc,iui))
        pack              = Code.Module("pack%s_I%s"%(tc,iui))
        instruction       = tP["localReadInstruction"]
        numOffsets        = instruction.numOffsets
        blockWidth        = instruction.blockWidth
        offsetMultiplier  = 1 # instruction.offsetMultiplier
        valuIdx           = 0
        numVectorsPerTile = (kernel["ThreadTile%u"%tile01]//kernel["VectorWidth"])
        numReadsPerVector = (kernel["VectorWidth"] * tP["bpe"]) // (blockWidth*4) # bytes/register

        for vIdx in range(0, numVectorsPerTile):
            for rIdx in range(0, numReadsPerVector):
                localReadCode = imod.addCode (Code.Module("LocalRead%s Valu%u"%(tc,valuIdx)))
                paramList     = []
                destVgpr      = vgpr("Valu%s_X%u_I%u+%u"%(tc, bufferIdx, iui, valuIdx), blockWidth)

                paramList.append(destVgpr)
                paramList.append(vgpr("LocalReadAddr%s"%tc))

                for oIdx in range(0, numOffsets):
                    paramList.append(((rIdx*blockWidth + kernel["SubGroup%u"%tile01] * (vIdx*numOffsets+oIdx)*kernel["VectorWidth"] \
                      + tP["localReadOffset"]) * tP["bpe"] + tP["localReadSwapByteOffset"]) // offsetMultiplier)
                    # print("Debug: Matrix{}, rIdx offset {}, vIdx offset {}, bpe {}, net offset {}".format( \
                    #     tP["tensorChar"], \
                    #     rIdx * blockWidth, \
                    #     kernel["SubGroup%u" % tP["tensorIdx"]] * (vIdx * numOffsets + oIdx) * kernel["VectorWidth"] + tP["localReadOffset"], \
                    #     tP["bpe"], \
                    #     paramList[-1]))
                paramTuple = tuple(paramList)
                comment = "L -> Reg lro=%d swapByteOffset=%u ti=%u vIdx=%u rIdx=%u oIdx=%u buffer=%u iui=%u"\
                    %(tP["localReadOffset"],tP["localReadSwapByteOffset"],kernel["SubGroup%u"%tile01], vIdx, rIdx, oIdx, bufferIdx, iui)
                localReadCode.addCode(Code.LocalReadInst(instruction.IssueLatency,False,instruction.toCodeInst(paramTuple), comment))
                valuIdx += blockWidth

                # TODO - handle vector-load
                tmpSgpr = writer.getTmpSgpr(1).idx()
                if writer.db["CheckValue1%s" % tc]:
                    dbgVgpr = destVgpr
                    dbgVgprList = destVgpr.split("v[")
                    if len(dbgVgprList) == 1: # vIdx, no []
                        dbgVgpr = dbgVgprList[0]
                    else:
                        # We only check the first one now
                        # TODO: Handle vector, but need to take care the last one
                        dbgVgprList = (dbgVgprList[1].split("]")[0]).split(':')
                        dbgVgpr = "v[%s]"%dbgVgprList[0]

                    localReadCode.addInst("s_waitcnt lgkmcnt(0)", "CheckValue1 wait for LDS read")
                    if writer.archCaps["SeparateVscnt"]:
                        localReadCode.addInst( "s_waitcnt_vscnt", "null", "0", "")

                    if kernel["ProblemType"]["DataType"].isHalf():
                        localReadCode.addInst("s_mov_b32", sgpr(tmpSgpr), hex(0x3c003c00),"CheckValue1: FP16")   # packed 1s
                        localReadCode.addCode(writer.assert_eq( dbgVgpr, sgpr(tmpSgpr)))

                    elif kernel["ProblemType"]["DataType"].isBFloat16():
                        localReadCode.addInst("s_mov_b32", sgpr(tmpSgpr), hex(0x3f803f80),"CheckValue1: BF16")   # packed 1s
                        localReadCode.addCode(writer.assert_eq( dbgVgpr, sgpr(tmpSgpr)))

                    # TODO - Check if this works
                    if kernel["ProblemType"]["DataType"].isInt8():
                        localReadCode.addInst("s_mov_b32", sgpr(tmpSgpr), hex(0x01010101),"CheckValue1: INT8")   # packed 1s
                        localReadCode.addCode(writer.assert_eq( dbgVgpr, sgpr(tmpSgpr)))

                    # TODO - Check if this works
                    elif kernel["ProblemType"]["DataType"].isInt8x4():
                        localReadCode.addCode(writer.assert_eq( dbgVgpr, 1))

                    elif kernel["ProblemType"]["DataType"].isSingle():
                        localReadCode.addCode(writer.assert_eq( dbgVgpr, 1.0) )

        return imod, pack


class LocalReadMFMA(LocalRead):
    kernel = {"EnableMatrixInstruction": True}

    """
    Local Read: Do It A/B
    iui = Inner Unroll Idx
    epsi = expand pointer swap index. Only used for PAP
    """
    def __call__(self, writer, bufferIdx, iui, epsi, tP):
        kernel = writer.kernel

        writer.localReadDoCnt += 1

        tc               = tP["tensorChar"]
        imod = Code.Module("LocalReadDo%s_I%s" % (tc,iui))

        if tc == "A":
            lrvw = writer.lrvwA
            writer.localReadDoCntA += 1
        else:
            lrvw = writer.lrvwB
            writer.localReadDoCntB += 1
        tile01           = tP["tile01Idx"]
        instruction      = tP["localReadInstruction"]

        numOffsets       = instruction.numOffsets
        blockWidth       = instruction.blockWidth
        bpr              = writer.bpr
        vectorWidth      = kernel["VectorWidth%s"%tc]
        vectorWidthA     = kernel["VectorWidthA"]
        vectorWidthB     = kernel["VectorWidthB"]
        MIWaveGroupShape = [ kernel["MatrixInstM"] * kernel["MatrixInstBM"] * kernel["MIWaveGroup"][0] * vectorWidthA, \
                             kernel["MatrixInstN"] * kernel["MatrixInstBN"] * kernel["MIWaveGroup"][1] * vectorWidthB]

        LdsPad           = kernel["LdsPad%s"%tc] if kernel["LdsBlockSizePerPad%s"%tc] == 0 else 0
        tileStride       = 1
        UnrollStride     = kernel["MacroTile%s" % tc] + LdsPad
        if kernel["UnrollMajorLDS%s" % tc]:
            tileStride   = kernel["_DepthULds"] + LdsPad
            UnrollStride = 1

        numVectorsPerTile = kernel["MIWaveTile"][tile01] // vectorWidth
        numReadsPerVector = vectorWidth if kernel["UnrollMajorLDS%s"%tc] else (vectorWidth * tP["bpe"]) // int(blockWidth * 4)
        numReadsPerUnroll = ceil(tP["bpe"] * lrvw / int(blockWidth * 4)) if kernel["UnrollMajorLDS%s"%tc] else kernel["MIInputPerThread"] # bytes/register

        numVgpr  = int(ceil(blockWidth))
        lrvwTile = writer.lrvwTileA if tc == "A" else writer.lrvwTileB
        numElementPerRead = int(blockWidth * 4) // tP['bpe'] // lrvwTile

        # pack register
        pack     = Code.Module("pack%s_I%s"%(tc,iui))
        hasEccHalf = writer.archCaps["HasEccHalf"]
        needPack = writer.needPackA if tc == "A" else writer.needPackB
        if needPack and (not kernel["VgprForLocalReadPacking"]):
            # allcate tmp vgpr only for no VgprForLocalReadPacking case
            # No ECC 0.25: pack one time 0x00ff00ff | (0x00ff00ff << 8)
            packTimesPerVgpr = (int(1/blockWidth) - 1) if hasEccHalf else 1
            tmpVgprIdx = writer.vgprPool.checkOut(writer.numVgprValuAPerBlock*writer.numReadsIterCoalescedA*packTimesPerVgpr if tc == 'A' \
                else writer.numVgprValuBPerBlock*writer.numReadsIterCoalescedB*packTimesPerVgpr, "local read pack")
            pack.addTempVgpr(tmpVgprIdx) # important, add to pack Module for later CheckIn

        valufIdx = 0
        for vIdx in range(0, numVectorsPerTile):
            for eIdx in range(0, numReadsPerVector):
                valuiIdx = int(valufIdx)
                localReadCode = imod.addCode (Code.Module("LocalRead%s Valu%u"%(tc,valuiIdx)))
                if needPack:
                    packCode = pack.addCode (Code.Module("packCode"))
                for rIdx in range(0, numReadsPerUnroll):
                    valuiIdx = int(valufIdx)
                    baseVgprStr = "Valu%s_X%u_I%u"%(tc, bufferIdx, iui)
                    baseLRVgpr = vgpr(baseVgprStr + "+%u"%(valuiIdx), numVgpr)
                    destVgpr = baseLRVgpr

                    highBitsForHalf = (blockWidth == 0.5) and ((rIdx % 2) == 1) # rIdx = 1
                    isHigh8Bits  = (blockWidth == 0.25) and ( ((rIdx % 4) % 2) == 1) # 1,3
                    isHigh16Bits = (blockWidth == 0.25) and ( ((rIdx % 4) //2) == 1) # 2,3

                    if needPack:
                        if lrvwTile > 1:
                            # wider local read + v_perm
                            # use allocated vgpr
                            highBitsForHalf = 0
                            isHigh8Bits = 0
                            isHigh16Bits = 0
                            numElemPerReg = bpr//tP["bpe"]
                            destVgpr = vgpr(baseVgprStr + "_D%u+%u"%(rIdx%(kernel["MIInputPerThread"]), vIdx*numVgpr), numVgpr)

                            # use wider local read + v_perm_b32
                            if rIdx == numReadsPerUnroll-1:
                                for i in range(0, numVgpr):
                                    # convert from [tile][MiInputPerThread][vector] to [tile][vector][MiInputPerThread]
                                    vgprIdx = (vIdx*numVgpr+i)*tP["bpe"]*kernel["MIInputPerThread"]//bpr*min(bpr//tP["bpe"],vectorWidth)
                                    vgprOffset = 0
                                    numElemPerRegIter = 2 if tP["bpe"] == 1 and vectorWidth <= 2 else numElemPerReg
                                    for vectorIdx in range(0, numElemPerRegIter):
                                        for elementIdx in range(0, tP["bpe"]*kernel["MIInputPerThread"]//bpr):
                                            src0BufIdx = elementIdx*numElemPerReg+1
                                            src1BufIdx = elementIdx*numElemPerReg
                                            srcVregOffset = i+vIdx*numVgpr
                                            destPack = vgpr(baseVgprStr + "+%u"%(vgprIdx+vgprOffset))
                                            src0Pack = vgpr(baseVgprStr + "_D%u+%u"%(src0BufIdx, srcVregOffset))
                                            src1Pack = vgpr(baseVgprStr + "_D%u+%u"%(src1BufIdx, srcVregOffset))
                                            src2Sgpr = sgpr("PackKForV%u"%vectorIdx)
                                            packCode.addInst("v_perm_b32", destPack, src0Pack, src1Pack, src2Sgpr, "select K=%u%u for vector=%u"%(src1BufIdx,  src0BufIdx, vectorIdx))
                                            if tP["bpe"] == 1:
                                                # run v_perm_b32 one more time for byte data
                                                src0BufIdx = elementIdx*numElemPerReg+3
                                                src1BufIdx = elementIdx*numElemPerReg+2
                                                destTmpVgpr = vgpr("PackTemp")
                                                src0Pack = vgpr(baseVgprStr + "_D%u+%u"%(src0BufIdx, srcVregOffset))
                                                src1Pack = vgpr(baseVgprStr + "_D%u+%u"%(src1BufIdx, srcVregOffset))
                                                packCode.addInst("v_perm_b32", destTmpVgpr, src0Pack, src1Pack, src2Sgpr, "select K=%u%u for vector=%u"%(src1BufIdx,  src0BufIdx, vectorIdx))
                                                packCode.addInst("_v_lshl_or_b32", destPack, destTmpVgpr, "0x10", destPack, "pack two half Vgpr to one Vgpr")

                                            vgprOffset += 1

                        else:
                            # bpe local read + or packing
                            if kernel["VgprForLocalReadPacking"]:
                                # use allocated vgpr
                                highVgprBase = baseVgprStr
                                highVgprHalf = vgpr(highVgprBase + "_D%u+%u"%(rIdx%2, valuiIdx), numVgpr)
                                highVgpr8Bits = vgpr(highVgprBase + "_D%u+%u"%(rIdx%4, valuiIdx), numVgpr)
                                lowVgpr = vgpr(highVgprBase + "_D%u+%u"%((rIdx%4)-1, valuiIdx), numVgpr) if isHigh16Bits else baseLRVgpr
                            else:
                                # use tmp vgpr
                                highVgprHalf = vgpr(tmpVgprIdx)
                                highVgpr8Bits = vgpr(tmpVgprIdx)
                                lowVgpr = vgpr(tmpVgprIdx-1) if isHigh16Bits else baseLRVgpr

                            if hasEccHalf: # ECC pack
                                # pack for ECC blockWidth 0.5 type
                                if needPack and highBitsForHalf:
                                    highVgpr = highVgprHalf
                                    packCode.addInst("v_or_b32", destVgpr, destVgpr, highVgpr, "pack two half Vgpr to one Vgpr")
                                    destVgpr = highVgpr

                                # pack for ECC blockwidth 0.25 type
                                if isHigh8Bits or isHigh16Bits:
                                    highVgpr = highVgpr8Bits
                                    destVgpr = highVgpr
                                if isHigh8Bits:
                                    packCode.addInst("_v_lshl_or_b32", lowVgpr, highVgpr, "0x8", lowVgpr, "pack two int8 Vgpr to one half Vgpr")
                                    if isHigh16Bits:
                                        packCode.addInst("v_or_b32", baseLRVgpr, baseLRVgpr, lowVgpr, "pack two half Vgpr to one Vgpr")

                                if (not kernel["VgprForLocalReadPacking"]) and (highBitsForHalf or isHigh8Bits or isHigh16Bits):
                                    tmpVgprIdx += 1

                            else: # no ECC pack
                                # pack for No ECC blockwidth 0.25 type
                                if isHigh8Bits:
                                    highVgpr = vgpr(tmpVgprIdx)
                                    destVgpr = highVgpr
                                if isHigh8Bits and isHigh16Bits:
                                    packCode.addInst("_v_lshl_or_b32", baseLRVgpr, highVgpr, "0x8", baseLRVgpr, "pack two int8x2 Vgpr to one Vgpr")
                                    tmpVgprIdx += 1

                    valufIdx += blockWidth

                    # load read instrution
                    paramList = []
                    paramList.append(destVgpr)
                    paramList.append(vgpr("LocalReadAddr%s"%tc))

                    for oIdx in range(0, numOffsets):
                        localReadOffset = tP["localReadOffset"]
                        tileStride2 = tileStride
                        localReadOffsetDiv = 0
                        if kernel["DirectToLds%s" % tc] and kernel["ThreadSeparateGlobalRead%c"%tc] and kernel["UnrollMajorLDS%s" % tc]:
                          ldsWidth = tileStride
                          ldsWidth -= LdsPad # do not include LdsPad
                          lrdOffsetMod = ldsWidth // (kernel["ThreadSeparateGlobalRead%c"%tc]*2)
                          tileStride2 //= (kernel["ThreadSeparateGlobalRead%c"%tc]*2) # change tileStride only for eIdx
                          localReadOffsetDiv = localReadOffset // lrdOffsetMod
                          localReadOffset = localReadOffset % lrdOffsetMod # keep only mod of lrdOffsetMod
                        offset_val = eIdx * tileStride2 + (vIdx * numOffsets+oIdx) * MIWaveGroupShape[tile01] * tileStride
                        # normal case
                        offset_val = (rIdx * numElementPerRead * UnrollStride + offset_val + localReadOffset) * tP["bpe"]
                        if localReadOffsetDiv > 0:
                          # TSGR special conversion
                          # Multiply BlockSize for each lrdOffsetMod
                          MblockSizePerLoad = (kernel["WavefrontSize"] * kernel["GlobalLoadVectorWidth%c"%tc]) // ldsWidth
                          offset_val += ((MblockSizePerLoad * lrdOffsetMod * tP["bpe"]) * localReadOffsetDiv)
                        if (kernel["LdsBlockSizePerPad%s"%tc] != 0) and (kernel["LdsPad%s"%tc] != 0):
                            offset_val = offset_val + (offset_val // kernel["LdsBlockSizePerPad%s"%tc]) * kernel["LdsPad%s"%tc] * tP["bpe"]
                        offset_val = offset_val + tP["localReadSwapByteOffset"]
                        if kernel["DirectToLds%s" % tc]:
                          # another address conversion for DirectToLds + NumLoadsCoalesced > 1
                          dummy, offset_val = writer.lraOffsetConversionForDTLandNLC(kernel, tP, offset_val)

                        paramList.append(int(offset_val))

                    paramTuple = tuple(paramList)
                    comment = "L -> Reg lro=%d swapByteOffset=%u ti=%u vIdx=%u rIdx=%u oIdx=%u buffer=%u iui=%u" \
                            % (tP["localReadOffset"], tP["localReadSwapByteOffset"], MIWaveGroupShape[tile01], vIdx, rIdx, oIdx, bufferIdx, iui)

                    highBits = highBitsForHalf or isHigh16Bits
                    readToTempVgpr = ((highBitsForHalf or isHigh8Bits or isHigh16Bits) if writer.archCaps["HasEccHalf"] else isHigh8Bits) and (not kernel["VgprForLocalReadPacking"])
                    localReadCode.addCode(Code.LocalReadInst(instruction.IssueLatency,readToTempVgpr,instruction.toCodeInst(paramTuple, 0, highBits), comment))

                    # TODO - handle vector-load
                    tmpSgpr = writer.getTmpSgpr(1).idx()
                    if writer.db["CheckValue1%s"%tc] and not writer.inTailLoop:
                        dbgVgpr = destVgpr
                        dbgVgprList = destVgpr.split("v[")
                        if len(dbgVgprList) == 1: # vIdx, no []
                            dbgVgpr = dbgVgprList[0]
                        else:
                            # We only check the first one now
                            # TODO: Handle vector, but need to take care the last one
                            dbgVgprList = (dbgVgprList[1].split("]")[0]).split(':')
                            dbgVgpr = "v[%s]"%dbgVgprList[0]

                        localReadCode.addInst("s_waitcnt lgkmcnt(0)", "CheckValue1 wait for LDS read")
                        if writer.archCaps["SeparateVscnt"]:
                            localReadCode.addInst( "s_waitcnt_vscnt", "null", "0", "")

                        if kernel["ProblemType"]["DataType"].isHalf():
                            hexValue = hex(0x3c003c00)     # packed 1s
                            if needPack:
                                hexValue = hex(0x3c000000) if highBitsForHalf else hex(0x00003c00)
                            localReadCode.addInst("s_mov_b32", sgpr(tmpSgpr), hexValue,"CheckValue1: FP16")
                            localReadCode.addCode(writer.assert_eq( dbgVgpr, sgpr(tmpSgpr)))

                        elif kernel["ProblemType"]["DataType"].isBFloat16():
                            hexValue = hex(0x3f803f80)     # packed 1s
                            if needPack:
                                hexValue = hex(0x3f800000) if highBitsForHalf else hex(0x00003f80)
                            localReadCode.addInst("s_mov_b32", sgpr(tmpSgpr), hexValue,"CheckValue1: BF16")
                            localReadCode.addCode(writer.assert_eq( dbgVgpr, sgpr(tmpSgpr)))

                        if kernel["ProblemType"]["DataType"].isInt8():
                            if needPack:
                                hexValue = hex(0x00010000) if isHigh16Bits else hex(0x00000001)
                                localReadCode.addInst("s_mov_b32", sgpr(tmpSgpr), hexValue,"CheckValue1: INT8")
                                localReadCode.addCode(writer.assert_eq( dbgVgpr, sgpr(tmpSgpr)))

                        # TODO - Check if this works. But need this? MFMA would use INT8
                        elif kernel["ProblemType"]["DataType"].isInt8x4():
                            localReadCode.addInst("s_mov_b32", sgpr(tmpSgpr), hex(0x01010101),"CheckValue1: INT8x4")
                            localReadCode.addCode(writer.assert_eq( dbgVgpr, sgpr(tmpSgpr)))

                        elif kernel["ProblemType"]["DataType"].isSingle():
                            localReadCode.addCode(writer.assert_eq( dbgVgpr, 1.0) )

        # DTV case, do not return local read code. Return pack code only.
        if kernel["DirectToVgpr%s"%tc]:
          return Code.Module(), pack

        return imod, pack



################################################################################
#
# Copyright (C) 2016-2022 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

from .Common import tPrint, printExit, printWarning
from .Utils import roundUpToNearestMultiple
from .AsmUtils import inst, vgpr, sgpr
import traceback
from enum import Enum
from typing import Optional

################################################################################
# RegisterPool
# Debugging register performance problems:
# - Enable self.db["PrintRP"] to see messages as vgprPool state changes.
# - Search for 'overlow' to see when pool grows dynamically - typically this
#   indicates growth for temps or other cases.
# - checkIn, checkout take optional tag but this is not widely used in tensile.
# - checkout returns vgpr index that was returned - can search disasm to see where
#   this vgpr is used.
################################################################################
class RegisterPool:
  class Status(Enum):
    Unavailable = 0
    Available = 1
    InUse = 2
    AvailableForPreload = 3

  class Register:
    def __init__(self, status, tag):
      self.status = status
      self.tag = tag

    def __repr__(self) -> str:
      return f"({self.status}, {self.tag})"

  ########################################
  # Init
  # defaultPreventOverflow: control behavior of checkout and checkoutAligned when preventOverflow is not explicitly specificed.
  def __init__(self, size, type, defaultPreventOverflow, printRP=0):
    self.printRP=printRP
    self.type = type
    self.defaultPreventOverflow = defaultPreventOverflow
    self.pool = [self.Register(RegisterPool.Status.Unavailable, "init") for i in range(0,size)]

    # Map from register number -> size, deleted/overwritten when deallocated
    self.checkOutSize = {}

    # Map from tag -> size, not deleted.
    self.checkOutSizeCache = {}

    # First register that's part of the manually loaded kernargs
    self.kernargStart = None
    # Last+1 register that's part of the manually loaded kernargs
    self.kernargEnd = None

    # First register that's part of the preloaded kernargs
    self.preloadStart = None
    # Last+1 register that's part of the preloaded kernargs
    self.preloadEnd = None

    # Each of the preloaded kernargs' name and size
    self.preloadedKernargs = []
    # Each of the manually loaded kernargs' name and size
    self.selfLoadedKernargs = []


  @property
  def numKernargSGPRs(self):
    if self.kernargStart is None:
      assert self.kernargEnd is None
      return 0
    return self.kernargEnd - self.kernargStart

  @property
  def numPreloadSGPRs(self):
    if self.preloadStart is None:
      assert self.preloadEnd is None
      return 0
    return self.preloadEnd - self.preloadStart

  @property
  def kernargs(self):
    from itertools import chain
    return chain(self.preloadedKernargs, self.selfLoadedKernargs)

  ########################################
  # Adds registers to the pool so they can be used as temps
  # Convenience function that takes a range and returns it in string form
  def addRange(self, start, stop, tag=""):
    self.add(start, stop-start+1, tag)
    if (start == stop):
      return "%d"%(start)
    else:
      return "%d-%d" % (start, stop)

  ########################################
  # Adds registers to the pool so they can be used as temps
  # Add
  def add(self, start, size, tag="", newStatus=Status.Available):
    # reserve space
    if self.printRP:
      print("RP::add(%u..%u for '%s')"%(start,start+size-1,tag))
    newSize = start + size
    oldSize = len(self.pool)
    if newSize > oldSize:
      for i in range(0, newSize-oldSize):
        self.pool.append(self.Register(RegisterPool.Status.Unavailable,tag))
    # mark as available
    for i in range(start, start+size):
      if self.pool[i].status == RegisterPool.Status.Unavailable:
        self.pool[i].status = newStatus
        self.pool[i].tag = tag
      elif self.pool[i].status == RegisterPool.Status.Available:
        printWarning("RegisterPool::add(%u,%u) pool[%u](%s) already available" % (start, size, i, self.pool[i].tag))
      elif self.pool[i].status == RegisterPool.Status.InUse:
        printWarning("RegisterPool::add(%u,%u) pool[%u](%s) already in use" % (start, size, i, self.pool[i].tag))
      else:
        raise RuntimeError("RegisterPool::add(%u,%u) pool[%u](%s) = %s" % (start, size, i, self.pool[i].tag, self.pool[i].status))
    if self.printRP:
      print(self.state())
  ########################################
  # Remove
  # Removes registers from the pool so they cannot be subsequently allocated for tmps
  def remove(self, start, size, tag=""):
    if self.printRP:
      print("RP::remove(%u..%u) for %s"%(start,size-1,tag))
    # reserve space
    newSize = start + size
    oldSize = len(self.pool)
    if newSize > oldSize:
      printWarning("RegisterPool::remove(%u,%u) but poolSize=%u" % (start, size, oldSize))
    # mark as unavailable
    for i in range(start, start+size):
      if  self.pool[i].status == RegisterPool.Status.Available:
        self.pool[i].status = RegisterPool.Status.Unavailable
      elif self.pool[i].status == RegisterPool.Status.Unavailable:
        printWarning("RegisterPool::remove(%u,%u) pool[%u](%s) already unavailable" % (start, size, i, self.pool[i].tag))
      elif  self.pool[i].status == RegisterPool.Status.InUse:
        printWarning("RegisterPool::remove(%u,%u) pool[%u](%s) still in use" % (start, size, i, self.pool[i].tag))
      else:
        printExit("RegisterPool::remove(%u,%u) pool[%u](%s) = %s" % (start, size, i, self.pool[i].tag, self.pool[i].status))

  ########################################
  # Check Out
  def checkOut(self, size, tag="_untagged_", preventOverflow=-1):
    return self.checkOutAligned(size, 1, tag, preventOverflow)

  def isRangeAvailable(self, start, size, preventOverflow=-1, wantedStatus=Status.Available) -> bool:
    end = start + size
    if preventOverflow == -1:
      preventOverflow = self.defaultPreventOverflow

    if preventOverflow and end > len(self.pool):
      return False

    end = min(end, len(self.pool))

    for i in range(start, end):
      if self.pool[i].status != wantedStatus:
        return False

    return True


  def findFreeRange(
    self,
    size: int,
    alignment: int,
    preventOverflow: int=-1,
    wantedStatus: Status=Status.Available
  ) -> Optional[int]:
    """Find a free range of registers of a given size, alignment, and status from the register pool.

    Args:
        size: The required size of the register block.
        alignment: The alignment requirement for the block. Each block must start at an index
            that is a multiple of this value.
        preventOverflow: Flag to prevent overflow. If set to -1, it uses the default value.
        wantedStatus: The status of the block to be found.

    Returns:
        1. The starting index of the block if found.
        2. None if preventOverflow is set and no block is found.
        3. The start of the last available block rounded up to the alignment if preventOverflow is
            not set and no block is found.
    """
    if preventOverflow == -1:
      preventOverflow = self.defaultPreventOverflow

    left, right = 0, 0
    while right < len(self.pool):
      if self.pool[right].status == wantedStatus:
        blockSize = right - left + 1
        if blockSize >= size:
          return left
        right += 1
      else:
        left += alignment
        right = left

    return None if preventOverflow else left


  def checkOutAt(self, start, size, tag, preventOverflow, wantedStatus = Status.Available):
    if preventOverflow:
      assert start + size <= len(self.pool)

    assert self.isRangeAvailable(start, size, preventOverflow, wantedStatus=wantedStatus)

    end = start + size

    numToAdd = max(0, end - len(self.pool))
    if numToAdd > 0:
      self.add(len(self.pool), numToAdd)

    for i in range(start, end):
      assert self.pool[i].status == wantedStatus

    for i in range(start, end):
      self.pool[i].status = RegisterPool.Status.InUse
      self.pool[i].tag = tag
    self.checkOutSize[start] = size
    self.checkOutSizeCache[tag] = size

  def startOfLastAvailableBlock(self) -> int:
    """ Returns the index of the first available register in the highest-numbered free block of registers. """
    for i in range(len(self.pool)-1, 0, -1):
      if self.pool[i].status != RegisterPool.Status.Available:
        return i+1
    return len(self.pool)

  def checkOutAligned(self, size, alignment, tag="_untagged_aligned_", preventOverflow=-1, kernarg=False, preload=False):
    if preventOverflow == -1:
      preventOverflow = self.defaultPreventOverflow
    assert size > 0

    if kernarg:
      assert not preventOverflow
      entry = {"name": tag, "size": size}
      if preload:
        self.preloadedKernargs.append(entry)
      else:
        self.selfLoadedKernargs.append(entry)

    if preload:
      loc = self.findFreeRange(size, alignment, True, RegisterPool.Status.AvailableForPreload)
      if self.preloadStart is None:
        assert loc == 2, "Assume that preloaded kernargs start at s2"
        self.preloadStart = loc
      else:
        assert loc == self.preloadEnd

      self.preloadEnd = loc + size
      assert loc is not None
      self.checkOutAt(loc, size, tag, True, wantedStatus=RegisterPool.Status.AvailableForPreload)
      return loc

    if kernarg:
      assert not preventOverflow
      if self.kernargStart is None:
        loc = self.startOfLastAvailableBlock()
        loc = roundUpToNearestMultiple(loc, alignment)
        self.kernargStart = loc
      else:
        loc = roundUpToNearestMultiple(self.kernargEnd, alignment)

      self.kernargEnd = loc + size
      self.checkOutAt(loc, size, tag, preventOverflow)
      return loc

    loc = self.findFreeRange(size, alignment, preventOverflow)
    assert loc is not None
    self.checkOutAt(loc, size, tag, preventOverflow)
    return loc

  def initTmps(self, initValue, start=0, stop=-1):
    kStr = ""
    stop= len(self.pool) if stop== -1 or stop>len(self.pool) else stop+1
    for i in range(start, stop):
      #if self.type == 's':
      #  print i, self.pool[i].status
      if self.pool[i].status==RegisterPool.Status.Available:
        if self.type == 's':
          kStr += inst("s_mov_b32", sgpr(i), hex(initValue), "init tmp in pool")
        elif self.type == 'v':
          kStr += inst("v_mov_b32", vgpr(i), hex(initValue), "init tmp in pool")
        else:
          assert(0) # bad regpool type

    return kStr

  ########################################
  # Check In
  def checkIn(self, start):
    if start in self.checkOutSize:
      size = self.checkOutSize[start]
      for i in range(start, start+size):
        self.pool[i].status = RegisterPool.Status.Available
      self.checkOutSize.pop(start)
      if self.printRP:
        print("RP::checkIn('%s') @ %u +%u"%(self.pool[i].tag, start,size))
    else:
      if 0:
        traceback.print_stack(None)
        import pdb; pdb.set_trace()
      printWarning("RegisterPool::checkIn('%s',%s) but it was never checked out"%(self.pool[start].tag, start))
    #traceback.print_stack(None)

  ########################################
  # Size
  def size(self):
    return len(self.pool)


  ########################################
  # Number of available registers
  def available(self):
    numAvailable = 0
    for s in self.pool:
      if s.status == RegisterPool.Status.Available:
        numAvailable += 1
    return numAvailable

  ########################################
  # Size of registers of at least specified blockSize
  def availableBlock(self, blockSize, align):
    if blockSize ==0:
      blockSize = 1
    blocksAvail = 0
    consecAvailable = 0
    #for s in self.pool:
    for i in range(0, len(self.pool)):
      s = self.pool[i]
      if s.status == RegisterPool.Status.Available:
        if not (consecAvailable == 0 and i % align != 0):
          # do not increment if the first item is not aligned
          consecAvailable += 1
      else:
        blocksAvail += consecAvailable // blockSize
        consecAvailable = 0
    blocksAvail += consecAvailable // blockSize
    #print self.state()
    #print "available()=", self.available(), "availableBlock()=",maxAvailable
    return blocksAvail * blockSize

  def availableBlockAtEnd(self):
    availCnt = 0
    for s in reversed(self.pool):
      if s.status == RegisterPool.Status.Available:
        availCnt += 1
      else:
        break

    return availCnt


  ########################################
  def checkFinalState(self):
    for si in range(0,len(self.pool)):
      if self.pool[si].status == RegisterPool.Status.InUse:
        if self.printRP:
          print(self.state())
        raise RuntimeError("RegisterPool::checkFinalState: temp (%s, '%s') was never checked in." \
            %(si, self.pool[si].tag))
    tPrint(3, "total vgpr count: %u\n"%self.size())

  ########################################
  # State
  def state(self):
    stateStr = ""
    placeValues = [1000, 100, 10, 1]
    for placeValueIdx in range(1, len(placeValues)):
      placeValue = placeValues[placeValueIdx]
      priorPlaceValue = placeValues[placeValueIdx-1]
      if len(self.pool) >= placeValue:
        pvs = "" # place value string
        for i in range(0, len(self.pool)):
          if i % placeValue==0:
            pvs += "%u"%((i%priorPlaceValue)//placeValue)
          else:
            pvs += " "
        stateStr += pvs + "\n"
    for i in range(0, len(self.pool)):
      if self.pool[i].status == RegisterPool.Status.Unavailable:
        stateStr += "." # 'removed', this indicates a fixed assignment from "remove", ie a non-tmp allocation
      elif self.pool[i].status == RegisterPool.Status.Available:
        stateStr += "|" # Can be allocated
      elif self.pool[i].status == RegisterPool.Status.InUse:
        stateStr += "#" # Checked out
    return stateStr

  def stateDetailed(self):
    for index, register in enumerate(self.vgprPool.pool):
        print("%u: %s"%(index, register.tag))



#!/usr/bin/python

################################################################################
#
# Copyright (C) 2018-2022 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

# Create a test_py script for all *.yaml files in specified directory
# usage: create_tests.py TEST_DIR
# Run from the Tensile/Tests directory, output script goes in the TEST_DIR/test_TEST_DIR.py

# The directory containing the test script can be passed to pytest:
# PYTHONPATH=. py.test-3 --durations=0 -v Tensile/Tests/TEST_DIR/
from __future__ import print_function
import glob, sys, os

targetDir  = sys.argv[1] if len(sys.argv) > 1 else "."
targetFile = "%s/test_%s.py"%(targetDir,os.path.basename(targetDir))
print("info: writing test script to %s" % targetFile)
outfile = open(targetFile, "w" )
outfile.write("import Tensile.Tensile as Tensile\n\n")
for f in glob.glob("%s/*aml"%targetDir):
    baseName = os.path.basename(f)
    testName = os.path.splitext(baseName)[0]
    testName = testName.replace('.','_')
    if not testName.startswith("test_"):
        testName = "test_" + testName

    outfile.write ("def %s(tmpdir):\n" % (testName))
    outfile.write (' Tensile.Tensile([Tensile.TensileTestPath("%s"), tmpdir.strpath])\n\n' % (f))



################################################################################
#
# Copyright (C) 2022 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

from ..Component import Component, MAC
from ..DataType import DataType

class FMA_BF16_HPA(MAC):
    asmCaps = {"v_fma_f32": True}
    kernel = {"ProblemType": {"DataType": DataType(DataType.bfloat16),
                              "HighPrecisionAccumulate": True}}

    def __call__(self, writer, m, innerUnroll):
        kernel = writer.kernel
        kStr = self.commentHeader()
        priority = Component.Priority.find(writer)

        vars = {}
        vars["endLine"] = writer.endLine
        vars["m"] = m
        vars["ThreadTile0"] = kernel["ThreadTile0"]
        vars["ThreadTile0Half"] = kernel["ThreadTile0"] // 2

        for iui in range(0, innerUnroll):
            vars["iui"] = iui
            for blockA in range(kernel["ThreadTileA"]//2-1, -1, -1):
                vars["blockA"] = blockA
                kStr += "v_and_b32 v[vgprValuA_X{m}_I{iui}+{blockA}*2+1], 0xffff0000, v[vgprValuA_X{m}_I{iui}+{blockA}]{endLine}".format_map(vars)
                kStr += "v_lshlrev_b32 v[vgprValuA_X{m}_I{iui}+{blockA}*2], 16, v[vgprValuA_X{m}_I{iui}+{blockA}]{endLine}".format_map(vars)
            for blockB in range(kernel["ThreadTileB"]//2-1, -1, -1):
                vars["blockB"] = blockB
                kStr += "v_and_b32 v[vgprValuB_X{m}_I{iui}+{blockB}*2+1], 0xffff0000, v[vgprValuB_X{m}_I{iui}+{blockB}]{endLine}".format_map(vars)
                kStr += "v_lshlrev_b32 v[vgprValuB_X{m}_I{iui}+{blockB}*2], 16, v[vgprValuB_X{m}_I{iui}+{blockB}]{endLine}".format_map(vars)

        for block1 in range(0, kernel["ThreadTile1"]//2):
            vars["block1"] = block1
            for block0 in range(0, kernel["ThreadTile0"]//2):
                vars["block0"] = block0
                if kernel["ProblemType"]["HighPrecisionAccumulate"]:
                    # we treat HighPrecisionAccumulate as expanded packed math
                    for iui in range(0, innerUnroll):
                        vars["iui"] = iui

                        vars["blockA"] = block0 if writer.tPB["tile01Idx"] else block1
                        vars["blockB"] = block1 if writer.tPB["tile01Idx"] else block0

                        vars["aStr0"] = "v[vgprValuA_X{m}_I{iui}+{blockA}*2+0]".format_map(vars)
                        vars["aStr1"] = "v[vgprValuA_X{m}_I{iui}+{blockA}*2+1]".format_map(vars)
                        vars["bStr0"] = "v[vgprValuB_X{m}_I{iui}+{blockB}*2+0]".format_map(vars)
                        vars["bStr1"] = "v[vgprValuB_X{m}_I{iui}+{blockB}*2+1]".format_map(vars)

                        vars["cidx"] = block0*2 + block1*kernel["ThreadTile0"]*2 + 0
                        vars["cStr"] = "v[vgprValuC+{block0}*2+{block1}*{ThreadTile0}*2+0*2+0]".format_map(vars)
                        kStr += "v_fma_f32 {cStr}, {aStr0}, {bStr0}, {cStr} //ValuC[{cidx}]{endLine}".format_map(vars)

                        kStr += priority(writer, 1, "Raise priority while processing macs")

                        vars["aStr"] = vars["aStr1"] if writer.tPB["tile01Idx"] else vars["aStr0"]
                        vars["bStr"] = vars["bStr0"] if writer.tPB["tile01Idx"] else vars["bStr1"]
                        vars["cidx"] = block0*2 + block1*kernel["ThreadTile0"]*2 + 1
                        vars["cStr"] = "v[vgprValuC+{block0}*2+{block1}*{ThreadTile0}*2+0*2+1]".format_map(vars) # *2 b/c of fp32
                        kStr += "v_fma_f32 {cStr}, {aStr}, {bStr}, {cStr} //ValuC[{cidx}]{endLine}".format_map(vars)

                        vars["aStr"] = vars["aStr0"] if writer.tPB["tile01Idx"] else vars["aStr1"]
                        vars["bStr"] = vars["bStr1"] if writer.tPB["tile01Idx"] else vars["bStr0"]
                        vars["cidx"] = block0*2 + block1*kernel["ThreadTile0"]*2 + kernel["ThreadTile0"] + 0
                        vars["cStr"] = "v[vgprValuC+{block0}*2+{block1}*{ThreadTile0}*2+{ThreadTile0Half}*2+0]".format_map(vars)
                        kStr += "v_fma_f32 {cStr}, {aStr}, {bStr}, {cStr} //ValuC[{cidx}]{endLine}".format_map(vars)

                        vars["cidx"] = block0*2 + block1*kernel["ThreadTile0"]*2 + kernel["ThreadTile0"] + 1
                        vars["cStr"] = "v[vgprValuC+{block0}*2+{block1}*{ThreadTile0}*2+{ThreadTile0Half}*2+1]".format_map(vars)
                        kStr += "v_fma_f32 {cStr}, {aStr1}, {bStr1}, {cStr} //valuC[{cidx}]{endLine}".format_map(vars)
                        """
                        ignore this, not quite correct for mixed precision
                        D.f[31:16] = S0.f[31:16] * S1.f[31:16] + S2.f[31:16]
                        D.f[15:00] = S0.f[15:00] * S1.f[15:00] + S2.f[15:00]
                        C[0] = A[0]*B[0]+D[0]
                        C[1] = A[1]*B[1]+D[1]
                        """
                        #kStr += self.bomb(-13)

        kStr += priority(writer, 0, "Reset priority after macs")
        return kStr



import warnings

def showwarning(message, category, filename, lineno, file=None, line=None):
    msg = f"> {category.__name__}: {message}"
    if TENSILE_TERM_COLORS:
        msg = f"[yellow]{msg}[/yellow]"
    print(msg)

warnings.showwarning = showwarning

TENSILE_TERM_COLORS: bool = False
try:
    from rich import print as print
    TENSILE_TERM_COLORS = True
except ImportError:
    print = print


try:
    from yaml import CSafeLoader as yamlLoader
except ImportError:
    from yaml import SafeLoader as yamlLoader

try:
    from yaml import CSafeDumper as yamlDumper
except ImportError:
    from yaml import SafeDumper as yamlDumper


try:
    import joblib
except:
    warnings.warn("Missing dependency 'joblib', program will run without parallelism")
    joblib = None



################################################################################
#
# Copyright (C) 2020-2022 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

from ..Component import Component, MAC
from ..DataType import DataType

class FMA_F16_HPA_MAD_MIX_LDL(MAC):
    @staticmethod
    def asmCaps(caps):
        return (caps['v_mad_mix_f32'] or caps['v_fma_mix_f32']) \
            and not caps["v_dot2c_f32_f16"] \
            and not caps["v_dot2_f32_f16"]
    #archCaps = {}
    kernel = {"ProblemType": {"DataType": DataType(DataType.half),
                              "HighPrecisionAccumulate": True},
              "LocalDotLayout": lambda ldl: ldl > 1,
              "InnerUnroll": 2
             }

    def __call__(self, writer, m, innerUnroll):
        kernel = writer.kernel

        kStr = self.commentHeader()
        priority = Component.Priority.find(writer)

        vars = {}

        if writer.asmCaps["v_fma_mix_f32"]:
            vars["instruction"] = "v_fma_mix_f32"
        else:
            vars["instruction"] = "v_mad_mix_f32"

        vars["m"] = m
        vars["kernel"] = kernel
        vars["endLine"] = writer.endLine

        vars["ThreadTile0"] = kernel["ThreadTile0"]
        vars["ThreadTile1"] = kernel["ThreadTile1"]

        vars["Half_ThreadTile0"] = kernel["ThreadTile0"] // 2
        vars["Half_ThreadTile1"] = kernel["ThreadTile1"] // 2

        for blockB in range(0, kernel["ThreadTile1"]//2):
            for blockA in range(0, kernel["ThreadTile0"]//2):
                vars["blockA"] = blockA
                vars["blockB"] = blockB

                vars["aBase0"] = "vgprValuA_X{m}_I0".format_map(vars)
                vars["bBase0"] = "vgprValuB_X{m}_I0".format_map(vars)

                vars["aBase1"] = "vgprValuA_X{m}_I1".format_map(vars)
                vars["bBase1"] = "vgprValuB_X{m}_I1".format_map(vars)

                # we treat HighPrecisionAccumulate as expanded packed math
                vars["cIdxExpr"] = "{blockA}*2 + {blockB}*{ThreadTile0}*2 + 0*2 + 0".format_map(vars)
                vars["cidx"] = eval(vars["cIdxExpr"])

                vars["cStr"] = "v[vgprValuC + {cIdxExpr}]".format_map(vars) # *2 b/c of fp32

                vars["aStr"] = "v[{aBase0}+{blockA}]".format_map(vars)
                vars["bStr"] = "v[{bBase0}+{blockB}]".format_map(vars)

                kStr += "{instruction} {cStr}, {aStr}, {bStr}, {cStr} op_sel:[0,0,0] op_sel_hi:[1,1,0] //ValuC[{cidx}]{endLine}".format_map(vars)

                kStr += priority(writer, 1, "Raise priority while processing macs")

                kStr += "{instruction} {cStr}, {aStr}, {bStr}, {cStr} op_sel:[1,1,0] op_sel_hi:[1,1,0] //ValuC[{cidx}]{endLine}".format_map(vars)

                vars["cIdxExpr"] = "{blockA}*2 + {blockB}*{ThreadTile0}*2 + 0*2 + 1".format_map(vars)
                vars["cidx"] = eval(vars["cIdxExpr"])

                vars["cStr"] = "v[vgprValuC + {cIdxExpr}]".format_map(vars) # *2 b/c of fp32
                vars["aStr"] = "v[{aBase1}+{blockA}]".format_map(vars)
                vars["bStr"] = "v[{bBase0}+{blockB}]".format_map(vars)

                kStr += "{instruction} {cStr}, {aStr}, {bStr}, {cStr} op_sel:[0,0,0] op_sel_hi:[1,1,0] //ValuC[{cidx}]{endLine}".format_map(vars)
                kStr += "{instruction} {cStr}, {aStr}, {bStr}, {cStr} op_sel:[1,1,0] op_sel_hi:[1,1,0] //ValuC[{cidx}]{endLine}".format_map(vars)


                vars["cIdxExpr"] = "{blockA}*2 + {blockB}*{ThreadTile0}*2 + {Half_ThreadTile0}*2 + 0".format_map(vars)
                vars["cidx"] = eval(vars["cIdxExpr"])

                vars["cStr"] = "v[vgprValuC + {cIdxExpr}]".format_map(vars)
                vars["aStr"] = "v[{aBase0}+{blockA}]".format_map(vars)
                vars["bStr"] = "v[{bBase1}+{blockB}]".format_map(vars)

                kStr += "{instruction} {cStr}, {aStr}, {bStr}, {cStr} op_sel:[0,0,0] op_sel_hi:[1,1,0] //ValuC[{cidx}]{endLine}".format_map(vars)
                kStr += "{instruction} {cStr}, {aStr}, {bStr}, {cStr} op_sel:[1,1,0] op_sel_hi:[1,1,0] //ValuC[{cidx}]{endLine}".format_map(vars)

                vars["cidx"] = blockA*2 + blockB*kernel["ThreadTile0"]*2 + kernel["ThreadTile0"] + 1

                vars["cIdxExpr"] = "{blockA}*2 + {blockB}*{ThreadTile0}*2 + {Half_ThreadTile0}*2+1".format_map(vars)
                vars["cidx"] = eval(vars["cIdxExpr"])

                vars["cStr"] = "v[vgprValuC + {cIdxExpr}]".format_map(vars)
                vars["aStr"] = "v[{aBase1}+{blockA}]".format_map(vars)
                vars["bStr"] = "v[{bBase1}+{blockB}]".format_map(vars)

                kStr += "{instruction} {cStr}, {aStr}, {bStr}, {cStr} op_sel:[0,0,0] op_sel_hi:[1,1,0] //valuC[{cidx}]{endLine}".format_map(vars)
                kStr += "{instruction} {cStr}, {aStr}, {bStr}, {cStr} op_sel:[1,1,0] op_sel_hi:[1,1,0] //valuC[{cidx}]{endLine}".format_map(vars)
                #kStr += writer.bomb(-13)

        kStr += priority(writer, 0, "Reset priority after macs")

        return kStr


class FMA_F16_HPA_MAD_MIX(MAC):
    asmCaps = lambda caps: caps['v_mad_mix_f32'] or caps['v_fma_mix_f32']
    #archCaps = {}
    kernel = {"ProblemType": {"DataType": DataType(DataType.half),
                              "HighPrecisionAccumulate": True},
              "LocalDotLayout": 1
             }

    def __call__(self, writer, m, innerUnroll):
        kernel = writer.kernel

        kStr = self.commentHeader()
        priority = Component.Priority.find(writer)

        vars = {}

        if writer.asmCaps["v_fma_mix_f32"]:
            vars["instruction"] = "v_fma_mix_f32"
        else:
            vars["instruction"] = "v_mad_mix_f32"

        vars["m"] = m
        vars["kernel"] = kernel
        vars["endLine"] = writer.endLine

        vars["ThreadTile0"] = kernel["ThreadTile0"]
        vars["ThreadTile1"] = kernel["ThreadTile1"]

        vars["Half_ThreadTile0"] = kernel["ThreadTile0"] // 2
        vars["Half_ThreadTile1"] = kernel["ThreadTile1"] // 2

        for block1 in range(0, kernel["ThreadTile1"]//2):
            for block0 in range(0, kernel["ThreadTile0"]//2):
                for iui in range(0, innerUnroll):
                    vars["block0"] = block0
                    vars["block1"] = block1
                    vars["blockA"] = block0 if writer.tPA["tileIdx"] == 0 else block1
                    vars["blockB"] = block1 if writer.tPB["tileIdx"] != 0 else block0
                    vars["iui"] = iui

                    vars["aBase"] = "vgprValuA_X{m}_I{iui}".format_map(vars)
                    vars["bBase"] = "vgprValuB_X{m}_I{iui}".format_map(vars)

                    vars["cIdxExpr"] = "{block0}*2 + {block1}*{ThreadTile0}*2 + 0*2 + 0".format_map(vars)
                    vars["cidx"] = eval(vars["cIdxExpr"])

                    vars["cStr"] = "v[vgprValuC + {cIdxExpr}]".format_map(vars) # *2 b/c of fp32
                    vars["aStr"] = "v[{aBase}+{blockA}]".format_map(vars)
                    vars["bStr"] = "v[{bBase}+{blockB}]".format_map(vars)
                    kStr += "{instruction} {cStr}, {aStr}, {bStr}, {cStr} op_sel:[0,0,0] op_sel_hi:[1,1,0] //ValuC[{cidx}] iui={iui}{endLine}".format_map(vars)

                    kStr += priority(writer, 1, "Raise priority while processing macs")

                    vars["cIdxExpr"] = "{block0}*2 + {block1}*{ThreadTile0}*2 + 0*2 + 1".format_map(vars)
                    vars["cidx"] = eval(vars["cIdxExpr"])

                    vars["cStr"] = "v[vgprValuC + {cIdxExpr}]".format_map(vars) # *2 b/c of fp32
                    vars["opSel"] = "op_sel:[1,0,0]" if writer.tPA["tileIdx"] == 0 else "op_sel:[0,1,0]"
                    kStr += "{instruction} {cStr}, {aStr}, {bStr}, {cStr} {opSel} op_sel_hi:[1,1,0] //ValuC[{cidx}]{endLine}".format_map(vars)

                    vars["cIdxExpr"] = "{block0}*2 + {block1}*{ThreadTile0}*2 + {Half_ThreadTile0}*2 + 0".format_map(vars)
                    vars["cidx"] = eval(vars["cIdxExpr"])

                    vars["cStr"] = "v[vgprValuC+{cIdxExpr}]".format_map(vars)
                    vars["opSel"] = "op_sel:[0,1,0]" if writer.tPA["tileIdx"] == 0 else "op_sel:[1,0,0]"
                    kStr += "{instruction} {cStr}, {aStr}, {bStr}, {cStr} {opSel} op_sel_hi:[1,1,0] //ValuC[{cidx}]{endLine}".format_map(vars)

                    vars["cIdxExpr"] = "{block0}*2+{block1}*{ThreadTile0}*2+{Half_ThreadTile0}*2+1".format_map(vars)
                    vars["cidx"] = eval(vars["cIdxExpr"])

                    vars["cStr"] = "v[vgprValuC+{cIdxExpr}]".format_map(vars)
                    kStr += "{instruction} {cStr}, {aStr}, {bStr}, {cStr} op_sel:[1,1,0] op_sel_hi:[1,1,0] //ValuC[{cidx}]{endLine}".format_map(vars)

        kStr += priority(writer, 0, "Reset priority after macs")

        return kStr

class FMA_F16_DOT2(MAC):
    asmCaps = lambda caps: caps["v_dot2c_f32_f16"] or caps["v_dot2_f32_f16"]
    #archCaps = {}
    kernel = {"ProblemType": {"DataType": DataType(DataType.half),
                              "HighPrecisionAccumulate": True},
              "LocalDotLayout": lambda ldl: ldl > 1
             }

    def __call__(self, writer, m, innerUnroll):
        kernel = writer.kernel

        kStr = self.commentHeader()

        accumulate = writer.asmCaps["v_dot2c_f32_f16"]

        priority = Component.Priority.find(writer)

        vars = {}

        if accumulate:
            vars["instruction"] = "_v_dot2acc_f32_f16"
        else:
            vars["instruction"] = "v_dot2_f32_f16"

        vars["m"] = m
        vars["kernel"] = kernel
        vars["endLine"] = writer.endLine

        vars["ThreadTile0"] = kernel["ThreadTile0"]
        vars["ThreadTile1"] = kernel["ThreadTile1"]

        vars["Half_ThreadTile0"] = kernel["ThreadTile0"] // 2
        vars["Half_ThreadTile1"] = kernel["ThreadTile1"] // 2
        vars["cSrc"] = ""

        for blockB in range(0, kernel["ThreadTile1"]//2):
            for blockA in range(0, kernel["ThreadTile0"]//2):
                vars["blockA"] = blockA
                vars["blockB"] = blockB

                vars["aBase0"] = "vgprValuA_X{m}_I0".format_map(vars)
                vars["bBase0"] = "vgprValuB_X{m}_I0".format_map(vars)

                vars["aBase1"] = "vgprValuA_X{m}_I1".format_map(vars)
                vars["bBase1"] = "vgprValuB_X{m}_I1".format_map(vars)

                # we treat HighPrecisionAccumulate as expanded packed math
                vars["cIdxExpr"] = "{blockA}*2 + {blockB}*{ThreadTile0}*2 + 0*2 + 0".format_map(vars)
                vars["cidx"] = eval(vars["cIdxExpr"])

                vars["cStr"] = "v[vgprValuC + {cIdxExpr}]".format_map(vars) # *2 b/c of fp32
                vars["aStr"] = "v[{aBase0}+{blockA}]".format_map(vars)
                vars["bStr"] = "v[{bBase0}+{blockB}]".format_map(vars)

                if not accumulate:
                    vars["cSrc"] = ", {cStr}".format_map(vars)

                kStr += "{instruction} {cStr}, {aStr}, {bStr}{cSrc} //ValuC[{cidx}]{endLine}".format_map(vars)

                kStr += priority(writer, 1, "Raise priority while processing macs")

                vars["cIdxExpr"] = "{blockA}*2 + {blockB}*{ThreadTile0}*2 + 0*2 + 1".format_map(vars)
                vars["cidx"] = eval(vars["cIdxExpr"])

                vars["cStr"] = "v[vgprValuC + {cIdxExpr}]".format_map(vars)
                vars["aStr"] = "v[{aBase1}+{blockA}]".format_map(vars)
                vars["bStr"] = "v[{bBase0}+{blockB}]".format_map(vars)

                if not accumulate:
                    vars["cSrc"] = ", {cStr}".format_map(vars)

                kStr += "{instruction} {cStr}, {aStr}, {bStr}{cSrc} //ValuC[{cidx}]{endLine}".format_map(vars)

                vars["cIdxExpr"] = "{blockA}*2 + {blockB}*{ThreadTile0}*2 + {Half_ThreadTile0}*2 + 0".format_map(vars)
                vars["cidx"] = eval(vars["cIdxExpr"])

                vars["cStr"] = "v[vgprValuC + {cIdxExpr}]".format_map(vars)
                vars["aStr"] = "v[{aBase0}+{blockA}]".format_map(vars)
                vars["bStr"] = "v[{bBase1}+{blockB}]".format_map(vars)

                if not accumulate:
                    vars["cSrc"] = ", {cStr}".format_map(vars)

                kStr += "{instruction} {cStr}, {aStr}, {bStr}{cSrc} //ValuC[{cidx}]{endLine}".format_map(vars)

                vars["cIdxExpr"] = "{blockA}*2 + {blockB}*{ThreadTile0}*2 + {Half_ThreadTile0}*2 + 1".format_map(vars)
                vars["cidx"] = eval(vars["cIdxExpr"])

                vars["cStr"] = "v[vgprValuC + {cIdxExpr}]".format_map(vars)
                vars["aStr"] = "v[{aBase1}+{blockA}]".format_map(vars)
                vars["bStr"] = "v[{bBase1}+{blockB}]".format_map(vars)

                if not accumulate:
                    vars["cSrc"] = ", {cStr}".format_map(vars)

                kStr += "{instruction} {cStr}, {aStr}, {bStr}{cSrc} //ValuC[{cidx}]{endLine}".format_map(vars)
                #kStr += writer.bomb(-13)

        kStr += priority(writer, 0, "Reset priority after macs")

        return kStr



################################################################################
#
# Copyright (C) 2022 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################
from . import __version__
from . import LibraryIO
from .Common import tPrint,HR,defaultBenchmarkCommonParameters,defaultProblemType
from .DataType import DataType

import argparse
import os
import sys

def TensileLibLogicToYaml(userArgs):
    tPrint(1, "")
    tPrint(1, HR)
    tPrint(1, "#")
    tPrint(1, "#  TensileLibLogicToYaml Library v{}".format(__version__))

     # argument parsing and related setup
    argParser = argparse.ArgumentParser()

    argParser.add_argument("LibLogicFile", type=os.path.realpath,
                           help="Library logic file to be converted to tensile input yaml file")
    argParser.add_argument("SolutionIndex", type=int,help="Solution index from library logic File"
                           ,
                           default=None)
    argParser.add_argument("OutputYaml", type=os.path.realpath,
                           help="OutputYaml path where output tensile yaml files are placed")

    argParser.add_argument('--skipMI','-s', action="store_true", help="Skips the MatrixInstruction field in the tensile yaml file"
                                "i.e Thread Tile and Work Group parameters without MI",required=False)

    args = argParser.parse_args(userArgs)
    logicFilePath = args.LibLogicFile
    solutionIndex = args.SolutionIndex
    tensileYamlFile = args.OutputYaml

    tPrint(1, "#  Library Logic: {}".format(logicFilePath))
    tPrint(1, "#  Solution Index: {}".format(solutionIndex))
    tPrint(1, "#")
    tPrint(1, HR)
    tPrint(1, "")
    libYaml = LibraryIO.readYAML(logicFilePath)
    if libYaml == "":
     raise RuntimeError("Yaml file data is empty, read yaml file :{} failed".format(logicFilePath))

    #reads library logic file. AllsolutionStates=>has solution data for all solution index
    fields = LibraryIO.rawLibraryLogic(libYaml)
    (versionString, scheduleName,architectureName,deviceNames,problemTypeState,allsolutionStates,indexOrder,exactLogic,rangeLogic, otherFields) = fields

    #Extract the solution data for the user specified solution Index
    currentIndexSolution = allsolutionStates[solutionIndex]

    if currentIndexSolution == "":
      raise RuntimeError("Could not find the matching data for the solution index:{} from the library logic file, Try different solution index".format(solutionIndex))

    #Skip the MI calculation if 9 bit MI is not needed or MatrixInstruction field is disabled
    isMatrixInsEnabled = False
    if "EnableMatrixInstruction" in currentIndexSolution:
      isMatrixInsEnabled = currentIndexSolution["EnableMatrixInstruction"]

      if currentIndexSolution["EnableMatrixInstruction"] and currentIndexSolution["MatrixInstruction"]:
        isMatrixInsEnabled = True
      else:
        tPrint(1, "Matrix instruction is disabled skipping the matrix instruction parameter ..")

    tensileYamlFileData = []
    if args.skipMI != True and isMatrixInsEnabled:
      checkMacroTileThreadTileWorkGroupMatches(currentIndexSolution)
      MIInstruction9Bits = form9BitMIInstruction(currentIndexSolution)
    else:
      MIInstruction9Bits = "None"

    #Extract the problem type parameters
    currentIndexProblemType = currentIndexSolution["ProblemType"]

    #Iterate over problem type parameters and form problem type yaml data
    problemTypeYamlData = formProblemTypeYamlData(currentIndexProblemType,versionString)
    tensileYamlFileData.append(problemTypeYamlData)

    #Iterate over Fork parameters and form the Yaml data
    forkParameterYamlData = formForkParametersYamlData(currentIndexSolution, MIInstruction9Bits)
    tensileYamlFileData.append(forkParameterYamlData)

    #Forms the Library logic string
    problemSizeYamlData = formProblemSizeYamlData(exactLogic,solutionIndex,scheduleName,deviceNames,architectureName)
    tensileYamlFileData.append(problemSizeYamlData)

    #Write the Formed Yaml data into the Yaml File
    writeToTensileYamlFile(tensileYamlFile,tensileYamlFileData)

def formProblemTypeYamlData(currentIndexProblemType, versionString):
    problemTypeYamlData = []
    if len(currentIndexProblemType) == 0:
       raise RuntimeError("Length of problem Type Parameters is empty!!, Please re-check the library logic file !")

    # Form Global parameters section as well
    problemTypeYamlData.append("GlobalParameters:\n")
    problemTypeYamlData.append("  MinimumRequiredVersion: {}\n\n".format(versionString['MinimumRequiredVersion']))
    problemTypeYamlData.append("BenchmarkProblems:\n  -\n    - # ProblemType\n")

    for problemTypeKey, problemTypeValue in currentIndexProblemType.items():
     # always print DataType, DestDataType, ComputeDataType, HighPrecisionAccumulate, OperationType fields
      if problemTypeKey == "DataType" or problemTypeKey == "DestDataType" or problemTypeKey == "ComputeDataType":
         problemTypeYamlData.append("      {}: {}\n".format(problemTypeKey, DataType(problemTypeValue).toChar().lower()))
         continue
      if (problemTypeKey == "HighPrecisionAccumulate" or problemTypeKey == "OperationType" or
          problemTypeKey == "TransposeA" or problemTypeKey == "TransposeB" or problemTypeKey == "UseBeta" or problemTypeKey == "Batched"):
         problemTypeYamlData.append("      {}: {}\n".format(problemTypeKey, problemTypeValue))
         continue
      if problemTypeKey in defaultProblemType:
         if problemTypeValue != defaultProblemType[problemTypeKey]:
            problemTypeYamlData.append("      {}: {}\n".format(problemTypeKey, problemTypeValue))

    return ''.join(str(x) for x in problemTypeYamlData)

def formForkParametersYamlData(currentIndexSolution, MIInstruction9Bits):
    forkParametersYamlData = []
    forkParametersYamlData.append("    - # BenchmarkProblemSizeGroup\n")
    forkParametersYamlData.append("      InitialSolutionParameters:\n")
    forkParametersYamlData.append("      BenchmarkCommonParameters:\n")
    forkParametersYamlData.append("      ForkParameters:\n")

    for forkKey, forkValue in currentIndexSolution.items():
      if MIInstruction9Bits == "None":
        if forkKey == "ProblemType" or forkKey == "MatrixInstruction":
          tPrint(1, "Continuing Matrix Instructions for Non MI\n")
          continue

        if forkKey == "MACInstruction":
          if forkValue == "MAC": #This is a bug in the library logic file
            forkValue ="MAD"
          forkParametersYamlData.append("         - {}: {}\n".format(forkKey,[forkValue]))
          continue

      else:
        if forkKey == "ProblemType" or forkKey == "ThreadTile" or forkKey == "WorkGroup" or forkKey == "MACInstruction":
          continue
        if forkKey == "MatrixInstruction":
          forkParametersYamlData.append(MIInstruction9Bits)
          continue

      # Find the matching index for fork key name from list of dictionaries => defaultBenchmarkCommonParameters
      index =  next((i for i,d in enumerate(defaultBenchmarkCommonParameters) if forkKey in d), None)
      if index != None:
        forkValue = [forkValue] # convert to list
        if forkValue != defaultBenchmarkCommonParameters[index][forkKey]:
          forkParametersYamlData.append("         - {}: {}\n".format(forkKey, forkValue))

    return ''.join(str(x) for x in forkParametersYamlData)

def form9BitMIInstruction(currentSolutionState):
    MIBlock = currentSolutionState["MIBlock"]
    MIWaveTile =  currentSolutionState["MIWaveTile"]
    MIWaveGroup = currentSolutionState["MIWaveGroup"]

    if len(MIBlock) == 0 or len(MIWaveTile) == 0  or len(MIWaveGroup) == 0:
     raise RuntimeError("Length of MIBlock:{0}, MIWave Tile:{1},MIWaveGroup:{2} cannot be empty".format(len(MIBlock),len(MIWaveTile),len(MIWaveGroup)))

    MIBlock1 = [MIBlock[i] for i in (0,1,2,3,4)]
    MIBlock5bits = ','.join([str(item) for item in MIBlock1])
    MIWaveTile2Bits = ','.join([str(item) for item in MIWaveTile])
    MIWaveGroup2Bits = ','.join([str(item) for item in MIWaveGroup])

    if len(MIBlock5bits) == 0 or len(MIWaveTile2Bits) == 0  or len(MIWaveGroup2Bits) == 0:
       raise RuntimeError("Length of MIBlock5bits:{0}, MIWaveGroup2Bits:{1},MIWaveGroup2Bits:{2} cannot be empty".format(len(MIBlock5bits),len(MIWaveTile2Bits),len(MIWaveGroup2Bits)))

    MIInstruction9Bits = "         - MatrixInstruction:\n           - ["+ MIBlock5bits + "," + MIWaveTile2Bits + "," + MIWaveGroup2Bits + "]\n"

    return MIInstruction9Bits

def formProblemSizeYamlData(exactLogic,solutionIndex,scheduleName,deviceNames,architectureName):
    problemSizeYamlData =[]

    problemSizeYamlData.append("      BenchmarkJoinParameters:\n")
    problemSizeYamlData.append("      BenchmarkFinalParameters:\n")
    problemSizeYamlData.append("         - ProblemSizes:\n")
    #Form the problem Size
    for (size, mapping) in exactLogic:
      if mapping[0] == solutionIndex:
        problemSizeYamlData.append("           - Exact: %s             # Eff: %s  Solution Index: %s\n" % (size,mapping[1],mapping[0]))

     #Form final library logic string
    problemSizeYamlData.append("#########################################################################################\n")
    problemSizeYamlData.append("LibraryLogic:\n")
    problemSizeYamlData.append("    ScheduleName: %s\n"%scheduleName)
    problemSizeYamlData.append("    DeviceNames: %s\n"%deviceNames)
    problemSizeYamlData.append("    ArchitectureName: \"%s\"\n"%architectureName)

    return ''.join(str(x) for x in problemSizeYamlData)

def checkMacroTileThreadTileWorkGroupMatches(currentSolution):
    MIBlock = currentSolution["MIBlock"]
    MIWaveTile =  currentSolution["MIWaveTile"]
    MIWaveGroup = currentSolution["MIWaveGroup"]

    if MIBlock == "" or MIWaveTile == "" or MIWaveGroup == "":
      raise RuntimeError("Length of MIBlock:{0}, MIWaveTile:{1},MIWaveGroup:{2} cannot be empty,Check the library logic file !\n".format(len(MIBlock),len(MIWaveTile),len(MIWaveGroup)))

    if not currentSolution.keys() & {'WavefrontSize'}:
      waveFrontSize = 64
    else:
      waveFrontSize =   int(currentSolution["WavefrontSize"])

    TT0,TT1,MT0,MT1,WG0,WG1 = calculateThreadTileMacroTileWorkGroupParameters(MIBlock,MIWaveTile,MIWaveGroup,waveFrontSize)

    if not currentSolution.keys() & {'MacroTile0','MacroTile1','ThreadTile','WorkGroup','MatrixInstM','MatrixInstN'}:
      raise RuntimeError("one or more of these fields MacroTile0,MacroTile1,ThreadTile, WorkGroup,MatrixInstM,MatrixInstN is missing in the library logic file!! ..\n")

    if MT0 != currentSolution["MacroTile0"]:
      raise RuntimeError("Macro Tile0 {0} does not match LibLogic value {1}".format(MT0, currentSolution["MacroTile0"]))

    if MT1 !=  currentSolution["MacroTile1"]:
      raise RuntimeError("Macro Tile1 {0} does not match LibLogic value {1}".format(MT1, currentSolution["MacroTile1"]))

    if TT0 != int(currentSolution["ThreadTile"][0]):
      raise RuntimeError("ThreadTile0 {0} does not match LibLogic value {1}".format(TT0, currentSolution["ThreadTile"][0]))

    if TT1 !=  int(currentSolution["ThreadTile"][1]):
      raise RuntimeError("ThreadTile1 {0} does not match LibLogic value {1}".format(TT1, currentSolution["ThreadTile"][1]))

    if WG0 !=  int(currentSolution["WorkGroup"][0]):
      raise RuntimeError("WorkGroup0 {0} does not match LibLogic value {1}\n".format(WG0, currentSolution["WorkGroup"][0]))

    if WG1 !=  int(currentSolution["WorkGroup"][1]):
      raise RuntimeError("WorkGroup1 {0} does not match LibLogic value {1}".format(WG1, currentSolution["WorkGroup"][1]))

def calculateThreadTileMacroTileWorkGroupParameters(MIBlock,MIWaveTile,MIWaveGroup,waveFrontSize):
    #Matrix M -->MIBlock[0]
    #Matrix N -->MIBlock[1]
    TT0 = int(MIWaveTile[0])
    TT1 = int(MIWaveTile[1])*int(MIBlock[1])
    MT0 =  int(MIBlock[0])*int(MIBlock[4])*int(MIWaveTile[0])*int(MIWaveGroup[0])
    MT1 =  int(MIBlock[1])*int(MIWaveTile[1])*int(MIWaveGroup[1])
    WG0 =  int(MIBlock[0]) * int(MIBlock[4])*int(MIWaveGroup[0])
    WG1 = int(MIWaveGroup[0]) * int(MIWaveGroup[1])*waveFrontSize // int(WG0)

    return TT0,TT1,MT0,MT1,WG0,WG1

def writeToTensileYamlFile(tensileYamlFile, tensileYamlData):
    try:
      os.makedirs(os.path.dirname(tensileYamlFile), exist_ok=True)
      with open(tensileYamlFile, "w") as fileHandle:
        fileHandle.writelines("%s" % place for place in tensileYamlData)
        fileHandle.close()
        tPrint(1, "Successfully created the Tensile Input Yaml File {} from library logic file !!.\n".format(tensileYamlFile))
    except (OSError, IOError):
      tPrint(1, "Error: Creating file {} Please provide file name in this format <filename>.yaml.".format(tensileYamlFile))

def main():
    TensileLibLogicToYaml(sys.argv[1:])



################################################################################
#
# Copyright (C) 2020-2022 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

from abc  import ABC
from abc  import abstractmethod
from copy import deepcopy

class KernelWriterBase(ABC):

  def __init__(self):
    super().__init__()

    self.state = {}

    self.endLine = "\n"
    self.getGroupIdStr = "hc_get_group_id"
    self.getNumGroupsStr = "hc_get_num_groups"
    self.getLocalIdStr = "hc_get_workitem_id"
    self.getGlobalIdStr = "hc_get_workitem_absolute_id"
    self.sharedDeclStr = "__shared__ "
    self.sharedPtrStr = ""
    self.globalPtrStr = ""
    self.syncStr = "__syncthreads();"
    self.fenceStr = self.syncStr
    self.macFStr = "fmaf"
    self.macDStr = "fma"
    self.int64Str = "int64_t"
    self.uint64Str = "uint64_t"
    self.vectorComponents = ["x", "y", "z", "w"]
    self.atomicCasStr = "atomicCAS"
    self.volatileStr = ""
    self.deviceFunctionStr = "__device__ "


  def extractIndices(self, extractFrom, varPrefix, indices):
    kStr = ""
    for (i, index) in enumerate(indices):
      kStr += "  unsigned int " + varPrefix + self.indexChars[index] + " = ( " + extractFrom
      for j in reversed(list(range(i+1, len(indices)))):
        index2 = indices[j]
        kStr += " / size" + self.indexChars[index2]
      kStr += ")"

      if len(indices) > 1:
        kStr += " % size" + self.indexChars[index]
      kStr += ";" + self.endLine

    return kStr


  @abstractmethod
  def getKernelName(self):
    pass


  @abstractmethod
  def getHeaderFileString(self):
    pass


  @abstractmethod
  def getSourceFileString(self):
    pass


  ##########################
  # make class look like dict
  def keys(self):
    return list(self.state.keys())

  def __len__(self):
    return len(self.state)

  def __iter__(self):
    return iter(self.state)

  def __getitem__(self, key):
    return self.state[key]

  def __setitem__(self, key, value):
    self.state[key] = value

  def __str__(self):
    return self.getKernelName()

  def __repr__(self):
    return self.__str__()

  def getAttributes(self):
    return deepcopy(self.state)

  def __hash__(self):
    return hash(str(self))

  def __eq__(self, other):
    return isinstance(other, KernelWriterBase) and str(self) == str(other)

  def __ne__(self, other):
    result = self.__eq__(other)
    if result is NotImplemented:
      return result
    return not result



################################################################################
#
# Copyright (C) 2020-2022 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

from __future__ import print_function
import os
import sys
import argparse
from Tensile.Utilities.ConditionalImports import yamlLoader, yamlDumper

HR = "################################################################################"

################################################################################
# Print Debug
################################################################################

def printWarning(message):
  print("Tensile::WARNING: %s" % message)
  sys.stdout.flush()

def printExit(message):
  print("Tensile::FATAL: %s" % message)
  sys.stdout.flush()
  sys.exit(-1)

try:
  import yaml
except ImportError:
  printExit("You must install PyYAML to use Tensile (to parse config files). See http://pyyaml.org/wiki/PyYAML for installation instructions.")

def ensurePath( path ):
  if not os.path.exists(path):
    os.makedirs(path)
  return path

################################################################################
# Library Logic Container
################################################################################
class LibraryLogic:

  def __init__(self,filename=None):

    if filename is not None:
      print ("# Reading Library Logic: " + filename)
      try:
        stream = open(filename, "r")
      except IOError:
        printExit("Cannot open file: %s" % filename )
      data = yaml.load(stream, yamlLoader)

      if isinstance(data, list):

        length = len(data)

        if (length > 0):
          self.__set_versionString(data[0]["MinimumRequiredVersion"])
        else:
          self.__set_versionString(None)

        if (length > 1):
          self.__set_scheduleName(data[1])
        else:
          self.__set_scheduleName(None)

        if (length > 2):
          self.__set_architectureName(data[2])
        else:
          self.__set_architectureName(None)

        if (length > 3):
          self.__set_deviceNames(data[3])
        else:
          self.__set_deviceNames(None)

        if (length > 4):
          self.__set_problemType(data[4])
        else:
          self.__set_problemType(None)

        if (length > 5):
          self.__set_solutionStates(data[5])
        else:
          self.__set_solutionStates(None)

        if (length > 6):
          self.__set_indexOrder(data[6])
        else:
          self.__set_indexOrder(None)

        if (length > 7):
          exactData = data[7]
          exactList = list()
          for exact in exactData:
            size = exact[0]
            if (len(size) > 4):
              exactOut = [size[:4],exact[1]]
              exactList.append(exactOut)
            else:
              exactList.append(exact)
          self.__set_exactLogic(exactList)
        else:
          self.__set_exactLogic(None)

        if (length > 8):
          self.__set_rangeLogic(data[8])
        else:
          self.__set_rangeLogic(None)

      else:
        printExit("Invalid Logic file: %s" % filename)

      stream.close()

    else:
      self.__set_versionString(None)
      self.__set_scheduleName(None)
      self.__set_architectureName(None)
      self.__set_deviceNames(None)
      self.__set_problemType(None)
      self.__set_solutionStates(None)
      self.__set_indexOrder(None)
      self.__set_exactLogic(None)
      self.__set_rangeLogic(None)

  #versionString
  def __get_versionString(self):
    return self.__versionString

  def __set_versionString(self,value):
    self.__versionString = value

  versionString = property(__get_versionString,__set_versionString)

  #scheduleName
  def __get_scheduleName(self):
    return self.__scheduleName

  def __set_scheduleName(self, value):
    self.__scheduleName = value

  scheduleName = property(__get_scheduleName,__set_scheduleName)

  #architectureName
  def __get_architectureName(self):
    return self.__architectureName

  def __set_architectureName(self,value):
    self.__architectureName = value

  architectureName = property(__get_architectureName,__set_architectureName)

  #deviceNames
  def __get_deviceNames(self):
    return self.__deviceNames

  def __set_deviceNames(self,value):
    self.__deviceNames = value

  deviceNames = property(__get_deviceNames,__set_deviceNames)


  #problemTypeState
  def __get_problemType(self):
    return self.__problemType

  def __set_problemType(self,value):
    self.__problemType = value

  problemType = property(__get_problemType,__set_problemType)

  #solutionStates
  def __get_solutionStates(self):
    return self.__solutionStates

  def __set_solutionStates(self,value):
    self.__solutionStates = value

  solutionStates = property(__get_solutionStates,__set_solutionStates)

  #indexOrder
  def __get_indexOrder(self):
    return self.__indexOrder

  def __set_indexOrder(self,value):
    self.__indexOrder = value

  indexOrder = property(__get_indexOrder,__set_indexOrder)


  #exactLogic
  def __get_exactLogic(self):
    return self.__exactLogic

  def __set_exactLogic(self,value):
    self.__exactLogic = value

  exactLogic = property(__get_exactLogic,__set_exactLogic)

  #rangeLogic
  def __get_rangeLogic(self):
    return self.__rangeLogic

  def __set_rangeLogic(self,value):
    self.__rangeLogic = value

  rangeLogic = property(__get_rangeLogic,__set_rangeLogic)

  def writeLibraryLogic(self,filename):

    data = []

    data.append({"MinimumRequiredVersion":self.versionString})
    data.append(self.scheduleName)
    data.append(self.architectureName)
    data.append(self.deviceNames)
    data.append(self.problemType)
    data.append(self.solutionStates)
    data.append(self.indexOrder)
    data.append(self.exactLogic)
    data.append(self.rangeLogic)

    if not data:
      printExit("No data to output")
    else:
      try:
        stream = open(filename, "w")
        yaml.dump(data, stream, yamlDumper, default_flow_style=None)
        stream.close()
      except IOError:
        printExit("Cannot open file: %s" % filename)


def MergeTensileLogicFiles(origionalLibraryLogic, exactLibraryLogic):

  mergedLibraryLogic = LibraryLogic()

  solutionList = origionalLibraryLogic.solutionStates
  solutionListExact = exactLibraryLogic.solutionStates

  indexKey = "SolutionIndex"
  # zero out solution indexes
  for solution in solutionList:
    solution[indexKey] = 0

  for solution in solutionListExact:
    solution[indexKey] = 0

  newSolutionOffset = len(solutionList)

  filterdSolutionExactList = []
  replicationMapping = {}
  idx = 0
  idxMapping = newSolutionOffset

  mergedSolutionList = []
  for solution in solutionList:
    mergedSolutionList.append(solution)

  # construct the mappings from the old exact kernal configurations
  # to their definitions in the merged files
  for solution in solutionListExact:
    if solution in solutionList:
      # if solution exists in the origional configuration the
      # its placement in the merged kernel configurations list
      # gets mapped to the pre-existing configuration
      idxOrg = solutionList.index(solution)
      replicationMapping[idx] = idxOrg

    else:
      filterdSolutionExactList.append(solution)
      # if the solution does not exist in the origional configurations
      # it gets mapped to the new offset
      replicationMapping[idx] = idxMapping
      mergedSolutionList.append(solution)
      idxMapping += 1

    idx += 1

  exactLogic = origionalLibraryLogic.exactLogic
  exactLogicExact = exactLibraryLogic.exactLogic

  filteredExactLogicExact = []

  # use the mapping from above to remap the exact logic
  # in the merged file
  for exact in exactLogicExact:
    # example exact entry [[123,124,1,123], [5, 4312.3]]
    # the first fiedl in [5, 4312.3] is the mapping to the
    # kernel configuration
    kernelIndex = exact[1][0]

    if kernelIndex in replicationMapping:
      exact[1][0] = replicationMapping[kernelIndex]

    filteredExactLogicExact.append(exact)

  sizeList, _ = zip(*filteredExactLogicExact)

  mergedExactLogic = []
  for logicMapping in exactLogic:
    if logicMapping[0] not in sizeList:
      mergedExactLogic.append(logicMapping)

  for logicMapping in filteredExactLogicExact:
    mergedExactLogic.append(logicMapping)

  # re index solutions
  index = 0
  for solution in mergedSolutionList:
    solution[indexKey] = index
    index += 1


  mergedLibraryLogic.versionString = origionalLibraryLogic.versionString
  mergedLibraryLogic.scheduleName = origionalLibraryLogic.scheduleName
  mergedLibraryLogic.architectureName = origionalLibraryLogic.architectureName
  mergedLibraryLogic.deviceNames = origionalLibraryLogic.deviceNames
  mergedLibraryLogic.problemType = origionalLibraryLogic.problemType
  mergedLibraryLogic.solutionStates = mergedSolutionList
  mergedLibraryLogic.indexOrder = origionalLibraryLogic.indexOrder
  mergedLibraryLogic.exactLogic = mergedExactLogic
  mergedLibraryLogic.rangeLogic  = origionalLibraryLogic.rangeLogic

  return mergedLibraryLogic


def ProcessMergeLogicFile(exactFileName, origionalFileName, outputFileName):

  _, fileName = os.path.split(exactFileName)

  print ("processing file: " + fileName)

  libraryLogic = LibraryLogic(origionalFileName)
  libraryLogicExact = LibraryLogic(exactFileName)

  mergedLibraryLogic = MergeTensileLogicFiles(libraryLogic,libraryLogicExact)

  mergedLibraryLogic.writeLibraryLogic(outputFileName)

def RunMergeTensileLogicFiles():

  print("")
  print(HR)
  print("# Merge Library Logic")
  print(HR)
  print("")

  ##############################################################################
  # Parse Command Line Arguments
  ##############################################################################

  argParser = argparse.ArgumentParser()
  argParser.add_argument("OrigionalLogicPath", help="Path to the origional LibraryLogic.yaml input files.")
  argParser.add_argument("ExactLogicPath", help="Path to the exact LibraryLogic.yaml input files.")
  argParser.add_argument("OutputPath", help="Where to write library files?")

  args = argParser.parse_args()

  origionalLogicPath = args.OrigionalLogicPath
  exactLogicPath = args.ExactLogicPath
  outputPath = args.OutputPath
  print ("Origional Logic Path: " + origionalLogicPath)
  print ("Exact Logic Path: " + exactLogicPath)
  print ("OutputPath: " + outputPath)

  print("")
  ensurePath(outputPath)
  if not os.path.exists(exactLogicPath):
    printExit("LogicPath %s doesn't exist" % exactLogicPath)

  exactLogicFiles = [os.path.join(exactLogicPath, f) for f in os.listdir(exactLogicPath) \
      if (os.path.isfile(os.path.join(exactLogicPath, f)) \
      and os.path.splitext(f)[1]==".yaml")]

  for exactLogicFilePath in exactLogicFiles:
    _, fileName = os.path.split(exactLogicFilePath)

    origionalLogicFilePath = os.path.join(origionalLogicPath, fileName)
    if os.path.isfile(origionalLogicFilePath):

      outputLogicFilePath = os.path.join(outputPath, fileName)

      try:
        ProcessMergeLogicFile(exactLogicFilePath, origionalLogicFilePath, outputLogicFilePath)
      except Exception as ex:
        print("Exception: {0}".format(ex))

    else:
      print ("# file does not exist in origional directory " + origionalLogicFilePath)


################################################################################
# Main
################################################################################
if __name__ == "__main__":
    RunMergeTensileLogicFiles()
